# -*- coding: utf-8 -*-
"""
Created on Thu Jul 17 07:45:32 2025

@author: 2016570
"""
from .transpilation.hls_timed import *
from .transpilation.astutils import *
from .base import Logic
import astunparse

class Py4hwGenerator:
    
    def generateSequentialCode(self, obj):
        tree = self.generateSequentialAST(self, obj)
        ast.fix_missing_locations(tree)
        return astunparse.unparse(tree)


    def generateSequentialAST(self, obj, useIfElseTree=False):
        
        if isinstance(obj, type):
            if issubclass(obj, Logic):
                return self.generateSequentialASTFromClass(obj, useIfElseTree)
            else:
                raise Exception(f'{type(obj)} not supported')
                
        elif isinstance(obj, Logic):
            return self.generateSequentialASTFromClass(obj.__class__, useIfElseTree)
        else:
            raise Exception(f'{type(obj)} not supported')

    def generateSequentialASTFromClass(self, clz, useIfElseTree):
        tree = get_class_ast(clz)

        tree = replaceConstants(tree, clz)

        tree = ScopedSelfAliasRewriter().visit(tree)
        tree = ReplacePrintByPass().visit(tree)
        tree = RemoveIrrelevantConditions().visit(tree)
        tree = SSARewriter().visit(tree)
        tree = removeUnusedAssignments(tree)

        tree = inline_yield_from(tree)
        
        #tree = transform_generator_to_fsm(tree)
        #tree  = yield_to_state(tree)
        
        run_node = getASTFromClassMethod(tree, 'run')
        
        build = CFGBuilder()
        build.build(run_node)
        
        removePassFromCFG(build)
        removeEmptyNodesFromCFG(build)
        removeIrrelevantConditionsFromCFG(build)
        removeUnusedTargetsInCFG(build)
        removeConstantControlFlowFromCFG(build)
        removeEmptyNodesFromCFG(build)
        yields = selectYieldingNodes(build)
        state = extractStates(build, yields)
        if (useIfElseTree):
            topif = createStateSelectionIfElseTree(state, yields)
        else:
            topif = createStateSelectionMatchCase(state, yields)
            topif = ReplaceIfElseTreeByMatchCase().visit(topif)
        tr = ClockMethodReplacer(topif)
        tree = tr.visit(tree)
        tr = KeepInitAndClockTransformer()
        tree = tr.visit(tree)
        return tree
    
        
    


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 19 12:44:39 2023

@author: dcr
"""

import math
import matplotlib.pyplot as plt
from matplotlib.patches import Arc
from matplotlib.patches import Ellipse

def drawLine(x0, y0, x1, y1, color):
    plt.plot([x0,x1], [y0,y1], color)

def drawRect(x0, y0, x1, y1, color):
    cx = (x0+x1)/2
    cy = (y0+y1)/2
    w = x1 - x0
    h = y1 - y0
    
    plt.plot([x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=color)
    
def drawEllipse(x0, y0, x1, y1, color):
    cx = (x0+x1)/2
    cy = (y0+y1)/2
    w = x1 - x0
    h = y1 - y0
    
    circle = Ellipse((cx, cy), w, h, color=color, fill=False)
    plt.gca().add_patch(circle)

def drawPoint(x, y, color):
    circle = plt.Circle((x, y), 3, color=color)
    plt.gca().add_patch(circle)
    


def drawArc(x0, y0, x1, y1, rx0, ry0, rx1, ry1, color):
    cx = (rx0+rx1)/2
    cy = (ry0+ry1)/2
    w = rx1-rx0
    h = ry1-ry0
    vx0 = (x0-cx)
    vy0 = (y0-cy)
    theta0 = math.atan2(vy0, vx0) * (180/math.pi)
    vx1 = (x1-cx)
    vy1 = (y1-cy)
    theta1 = math.atan2(vy1, vx1) * (180/math.pi)
    arc = Arc((cx,cy), w, h, 0, theta1, theta0, color=color)
    plt.gca().add_patch(arc)
    
    #print('theta0', theta0, 'theta1', theta1)
    
def tokenize_strings(line):
    i = 0
    ret = []
    
    s = ''
    instring = False
    
    while (i < len(line)):
        c = line[i]
        
        if (instring):
            if (c == '"'):
                instring = False
                ret.append(s)
                s = ''
            else:
                s += c
        else:
            if (c == '('):
                ret.append(c)
            elif (c == ')'):
                if (len(s) > 0):
                    ret.append(s)
                s = ''
                ret.append(c)
            elif (c == '"'):
                instring = True
            elif (c == ' '):
                if (len(s) > 0):
                    ret.append(s)
                s = ''
            else:
                s += c
            
        i += 1
        
        
    return ret

def listify_tokens(inlist, pos=0, level=0):
    ret = []
    while (pos < len(inlist)):
        token = inlist[pos]
        
        if (token == '('):
            subtokens, newpos = listify_tokens(inlist, pos+1, level+1)
            pos = newpos
            ret.append(subtokens)
        elif (token == ')'):
            return ret, pos+1
        else:
            ret.append(token)
            pos += 1
            
    return ret


class BDFObj():
    def __init__(self):
        self.child = []
        
    def draw(self, ox=0, oy=0, color='black'):
        #print('Draw:', type(self))
        for item in self.child:
            item.draw(ox, oy, color)
            
    def find(self, ty):
        for item in self.child:
            if (isinstance(item, ty)):
                return item
        raise Exception('no {} child '.format(ty))
        
    def findAll(self, ty):
        ret = []
        
        for item in self.child:
            if (isinstance(item, ty)):
                ret.append(item)
                
        return ret
    
    def bbox(self):
        # returns a bounding box of the object
        rx0 = math.inf
        ry0 = math.inf
        rx1 = -math.inf
        ry1 = -math.inf
        
        for item in self.child:
            ix0, iy0, ix1, iy1 = item.bbox()
            
            if (ix0 < rx0):
                rx0 = ix0
            if (iy0 < ry0):
                ry0 = iy0
            if (ix1 > rx1):
                rx1 = ix1
            if (iy1 > ry1):
                ry1 = iy1
                
        return rx0, ry0, rx1, ry1

class BDFTop(BDFObj):
    pass

class BDFPin(BDFObj):
    
    def draw(self, ox=0, oy=0, color='black'):
        r = self.find(BDFRect)
        pt = self.find(BDFPoint)
        
        #ox, oy =  pt.x, pt.y
        ox = r.rx
        oy = r.ry
        
        drw = self.find(BDFDrawing)
        
        for line in drw.findAll(BDFLine):
            line.draw(ox, oy, color)
            
        for text in self.findAll(BDFText):
            text.draw(ox, oy, color)
        
class BDFPort(BDFObj):
    pass
        
class BDFInput(BDFObj):
    pass

class BDFOutput(BDFObj):
    pass

class BDFConnector(BDFObj):
    def draw(self, ox=0, oy=0, color='black'):
        points = self.findAll(BDFPoint)
        x = [ox + pt.x for pt in points]
        y = [oy + pt.y for pt in points]
        plt.plot(x, y, color)
        
        for text in self.findAll(BDFText):
            text.draw(ox, oy, color)

class BDFSymbol(BDFObj):
    def draw(self, ox=0, oy=0, color='black'):
        r = self.find(BDFRect)
        ox, oy = r.rx, r.ry
        
        #r.draw('blue')
        
        for port in self.findAll(BDFPort):
            #pt = port.find(BDFPoint)
            
            port.draw(ox, oy, color)
            #for line in port.findAll(BDFLine):
            #    line.draw(ox, oy)
                
        drw = self.find(BDFDrawing)
        
        for line in drw.findAll(BDFObj):
            #print('draw', type(line))
            line.draw(ox, oy)

class BDFDrawing(BDFObj):
    pass

class BDFRect(BDFObj):
    
    def parse(self, line):
        line = line[5:-1]
        part = line.split()
        self.rx = int(part[0])
        self.ry = int(part[1])
        self.rw = int(part[2]) - self.rx
        self.rh = int(part[3]) - self.ry
        
    def draw(self, ox=0, oy=0, color='black'):
        plt.plot([self.rx,self.rx+self.rw,self.rx+self.rw,        self.rx,self.rx],
                 [self.ry,        self.ry,self.ry+self.rh,self.ry+self.rh,self.ry], color)
        
    def bbox(self):
        return self.rx, self.ry, self.rx + self.rw, self.ry + self.rh

class BDFRectangle(BDFObj):
    def parse(self, line):
        tokens = tokenize_strings(line)
        li = listify_tokens(tokens)[0]
        
        self.x0 = int(li[1][1])
        self.y0 = int(li[1][2])
        self.x1 = int(li[1][3])
        self.y1 = int(li[1][4])
        
    def draw(self, ox=0, oy=0, color='black'):
        drawRect(ox + self.x0, oy + self.y0, ox + self.x1, oy + self.y1, color)

class BDFCircle(BDFObj):
    
    def parse(self, line):
        line = line.split('(rect')[1]
        #print('CIRCLE LINE:', line)
        part = line[:-2].split()
        
        self.rx0 = int(part[0])
        self.ry0 = int(part[1])
        self.rx1 = int(part[2]) 
        self.ry1 = int(part[3]) 
        
    def draw(self, ox=0, oy=0, color='black'):
        drawEllipse(ox + self.rx0, oy + self.ry0, ox + self.rx1, oy + self.ry1, color)

class BDFPoint(BDFObj):
    
    def parse(self, line):
        line = line[3:-1]
        #print('parsing point', line)
        part = line.split()
        self.x = int(part[0])
        self.y = int(part[1])

class BDFLine(BDFObj):
    
    def parse(self, line):
        #print('parse', type(self))
        part = line.split(')')
        part = [x.split() for x in part]
        self.x0, self.y0 = int(part[0][2]), int(part[0][3])
        self.x1, self.y1 = int(part[1][1]), int(part[1][2])
    
    def draw(self, ox=0, oy=0, color='black'):
        plt.plot([ox+self.x0,ox+self.x1], [oy+self.y0,oy+self.y1], color)

class BDFText(BDFObj):
    
    
    def parse(self, line):
        # (text "Pis_7s[3]" (rect 400 464 446 476)(font "Arial" ))
        # (text "OUT" (rect 48 47 65 59)(font "Courier New" (bold))(invisible))
        #print('PARSE TEXT', line)
        tokens = tokenize_strings(line)
        li = listify_tokens(tokens)[0]
        
        assert(li[2][0] == 'rect')
        self.rx0 = int(li[2][1])
        self.ry0 = int(li[2][2])
        self.rx1 = int(li[2][3])
        self.ry1 = int(li[2][4])
        
        self.invisible = 'invisible' in li[-1]
        
        self.text = li[1]
        
    def draw(self, ox=0, oy=0, color='black'):
        #print('DRAW TEXT:', self.text)
        
        if not(self.invisible):
            ax = plt.gca()
            plt.text(ox+self.rx0, oy+self.ry0, self.text , fontsize=8) #, transform=ax.transAxes)
    
class BDFJunction(BDFObj):
    
    def parse(self, line):
        pass
    
    def draw(self, ox=0, oy=0, color='black'):
        pass

class BDFArc(BDFObj):
    def parse(self, line):
        # (arc (pt 8 61)(pt 8 51)(rect -13 40 20 73))
        line = line[4:-1]
        line = line.replace(')', ' ')

        part = line.split('(rect')
        
        pts = part[0].split()
        self.x0 = int(pts[1])
        self.y0 = int(pts[2])
        self.x1 = int(pts[4])
        self.y1 = int(pts[5])
        
        #print('ARC pt:', self.x0, self.y0, self.x1, self.y1)
        
        pr = part[1][:-1].split()
        
        self.rx0 = int(pr[0])
        self.ry0 = int(pr[1])
        self.rx1 = int(pr[2]) 
        self.ry1 = int(pr[3]) 
        
    def draw(self, ox=0, oy=0, color='black'):
        drawArc(ox + self.x0, oy + self.y0, ox + self.x1, oy + self.y1, 
                ox + self.rx0 , oy + self.ry0, ox + self.rx1 , oy + self.ry1, color=color)
        
        #drawPoint(ox + self.x0, oy + self.y0, color)
        #drawPoint(ox + self.x1, oy + self.y1, color)
        #drawLine(ox + self.x0, oy + self.y0, ox + self.x1, oy + self.y1, color)
    
class DrawBDF():
    def __init__(self):
        self.rx = 0
        self.ry = 0
    
    def readfile(self, filename):
        f = open(filename, 'r')
        lines = f.readlines()
        ret = []
        
        # remove comments
        incomment = False
        for line in lines:
            if not(incomment):
                if '/*' in line:
                    part = line.split('/*')
                    ret.append(part[0])
                    incomment = True
                else:
                    if (len(line) > 0):
                        ret.append(line)
            else:
                if '*/' in line:
                    part = line.split('*/')
                    ret.append(part[1])
                    incomment = False
                else:
                    pass
                
        lines = ret
        
        # remove \n
        ret = []
        for line in lines:
            parts = line.split('\n')
            for part in parts:
                if (len(part) > 0):
                    ret.append(part.strip()) 
                    
        return ret
    
    def load(self, filename):
        self.lines = self.readfile(filename)
        self.i =0
        
        self.top = BDFTop()
        self.parseObjects(self.top)
        
        #print('START DRAWING')
        #self.top.draw()

    def draw(self, figsize=None):
        self.top.draw()

        ar = 1
        if not(figsize is None):
            plt.rcParams['figure.figsize'] = figsize
            ar = figsize[0]/figsize[1]
            
        
        x0, y0, x1, y1 = self.top.bbox()

        w = x1-x0
        h = y1-y0

        #print('w', w)
        #print('h', h)

        maxaxis = max(w,h)

        x1 = x0 + maxaxis 
        y1 = y0 + maxaxis / ar

        plt.ylim(y1, y0)
        plt.xlim(x0, x1)

        plt.show()    
        
    def parseObjects(self, obj):
        #print('parse', type(obj))
        while (self.i < len(self.lines)):
            line = self.lines[self.i]
            self.i+=1
            
            #print('{}:{}'.format(self.i, line))
            
            if (line[:7] == '(header' ):
                pass
            elif (line[:4] == '(pin'):
                pin = BDFPin()
                self.parseObjects(pin)
                obj.child.append(pin)
            elif (line[:5] == '(port'):
                port = BDFPort()
                self.parseObjects(port)
                obj.child.append(port)
            elif (line == '(connector'):
                port = BDFConnector()
                self.parseObjects(port)
                obj.child.append(port)
            elif (line[:3] == '(pt'):
                pt = BDFPoint()
                pt.parse(line)
                obj.child.append(pt)
            elif (line[:7] == '(symbol'):
                symbol = BDFSymbol()
                self.parseObjects(symbol)
                obj.child.append(symbol)
            elif (line[:8] == '(drawing'):
                drawing = BDFDrawing()
                self.parseObjects(drawing)
                obj.child.append(drawing)
            elif (line[:5] == '(line'):
                lobj = BDFLine()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line[:4] == '(arc'):
                lobj = BDFArc()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line[:10] == '(rectangle'):
                lobj = BDFRectangle()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line[:5] == '(rect'):
                robj = BDFRect()
                robj.parse(line)
                obj.child.append(robj)
            elif (line[:7] == '(circle'):
                lobj = BDFCircle()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line == '(input)'):
                lobj = BDFInput()
                obj.child.append(lobj)
            elif (line == '(output)'):
                lobj = BDFInput()
                obj.child.append(lobj)
            elif (line[:5] == '(text'):
                lobj = BDFText()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line[:9] == '(junction'):
                lobj = BDFJunction()
                lobj.parse(line)
                obj.child.append(lobj)
            elif (line == ')'):
                #print('<<<<<', type(obj))
                return
            else:
                print('?? - ?? - ??', line)
                


# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 15:17:55 2025

@author: 2016570
"""
import py4hw

class PerformanceCounter(py4hw.Logic):
    
    def __init__(self, parent, name, reset, chipselect, address, read, write, writedata, readdata):
        
        super().__init__(parent, name)
        
        self.addIn('reset', reset)
        self.addIn('chipselect', chipselect)
        self.addIn('address', address)
        self.addIn('read', read)
        self.addIn('write', write)
        self.addIn('writedata', writedata)
        self.addOut('readdata', readdata)
        
        bit0 = self.wire('bit0')
        
        py4hw.Bit(self, 'add0', address, 0, bit0)
        
        
        counter = self.wire('counter', 64)
        one = self.wire('one')
        
        py4hw.Constant(self, 'one', 1, one)
        py4hw.Counter(self, 'counter', reset, one, counter)
        
        counter_low = self.wire('counter_low', 32)
        counter_high = self.wire('counter_high', 32)
        
        py4hw.Range(self, 'counter_low', counter, 31, 0, counter_low)
        py4hw.Range(self, 'counter_high', counter, 63, 32, counter_high)
        
        py4hw.Mux2(self, 'mux', bit0, counter_low, counter_high, readdata)
        
        
def generateVerilog():
    
    hw = py4hw.HWSystem()
    
    reset = hw.wire('reset')
    chipselect = hw.wire('chipselect')
    address = hw.wire('address', 2)
    read = hw.wire('read')
    write = hw.wire('write')
    writedata = hw.wire('writedata', 32)
    readdata = hw.wire('readdata', 32)
    
    perf = PerformanceCounter(hw, 'counter', reset, chipselect, address, read, write, writedata, readdata)
    
    rtl = py4hw.VerilogGenerator(perf)
    
    return rtl.getVerilogForHierarchy()
    


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  8 16:20:12 2024

@author: 2016570
"""

import py4hw

class cycloneiii_io_buf (py4hw.Logic):
    
    def __init__(self, parent, name, pin, pout, poe, bidir):
        
        super().__init__(parent, name)
        
        assert(poe.getWidth() == 1)
        
        self.pin = self.addOut('pin', pin)
        self.pout = self.addIn('pout', pout)
        self.poe = self.addIn('poe', poe)
        
        self.bidir = self.addInOut('bidir', bidir)
        
    def structureName(self):
        return 'cycloneiii_io_buf_{}'.format(self.bidir.getWidth())
    
    def propagate(self):
        pass
    
    def verilogBody(self):
        ret = ''
        
        ret += '    cycloneiii_io_ibuf   ibufa_0(.i(bidir),	.o(pin) );\n'
        
        ret += '    cycloneiii_io_obuf   obufa_0(.i(pout),	.o(bidir),	.oe(poe));\n'
    
        return ret


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 19 16:59:12 2023

@author: dcr
"""

def tokenize_strings(line):
    i = 0
    ret = []
    
    s = ''
    instring = False
    
    while (i < len(line)):
        c = line[i]
        
        if (instring):
            if (c == '"'):
                instring = False
                ret.append(s)
                s = ''
            else:
                s += c
        else:
            if (c == '('):
                ret.append(c)
            elif (c == ')'):
                if (len(s) > 0):
                    ret.append(s)
                s = ''
                ret.append(c)
            elif (c == '"'):
                instring = True
            elif (c == ' '):
                if (len(s) > 0):
                    ret.append(s)
                s = ''
            else:
                s += c
            
        i += 1
        
        
    return ret

def listify_tokens(inlist, pos=0, level=0):
    ret = []
    while (pos < len(inlist)):
        token = inlist[pos]
        
        if (token == '('):
            subtokens, newpos = listify_tokens(inlist, pos+1, level+1)
            pos = newpos
            ret.append(subtokens)
        elif (token == ')'):
            return ret, pos+1
        else:
            ret.append(token)
            pos += 1
            
    return ret

def startsWith(line, st):
    return (line[0:len(st)] == st)

def convertTransitionsToValues(lines, period):
    ret = []
    
    for line in lines:
        if (isinstance(line, VWFNode)):
            ret2 = convertTransitionsToValues(line.value, period)
            ret.extend(ret2)
            
        elif (startsWith(line, 'REPEAT')):
            part = line.split('=')
            repeat = part[1].strip()
        elif (startsWith(line, 'LEVEL')):
            part = line.split()
            level = part[1]
            assert(part[2] == 'FOR')
            dur = int(float(part[3]) / period)
            
            for i in range(dur):
                ret.append(level)
                
    return ret;

class VWFObj():
    def __init__(self):
        self.child = []
        
            
    def find(self, ty):
        for item in self.child:
            if (isinstance(item, ty)):
                return item
        raise Exception('no {} child '.format(ty))
        
    def findAll(self, ty):
        ret = []
        
        for item in self.child:
            if (isinstance(item, ty)):
                ret.append(item)
                
        return ret
    

class VWFTop(VWFObj):
    pass

class VWFHeader(VWFObj):
    pass

class VWFTimeBar(VWFObj):
    pass

class VWFDisplayLine(VWFObj):
    pass

class VWFSignal(VWFObj):
    
    def parse(self, line):
        line = line[len('SIGNAL('):-1] # trim 
        
        if (line[0] == '"'):
            line = line[1:-1] # remove quotes
                    
        self.name = line
        
class VWFTransitionList(VWFObj):
    
    def parse(self, line):
        line = line[len('TRANSITION_LIST('):-1] # trim 
        
        if (line[0] == '"'):
            line = line[1:-1] # remove quotes
                    
        self.name = line
    
class VWFNode(VWFObj):
    pass

class ConvertVWF():
    def __init__(self):
        self.verbose = True
    
    def readfile(self, filename):
        f = open(filename, 'r')
        lines = f.readlines()
        ret = []
        
        # remove comments
        incomment = False
        for line in lines:
            if not(incomment):
                if '/*' in line:
                    part = line.split('/*')
                    ret.append(part[0])
                    incomment = True
                else:
                    if (len(line) > 0):
                        ret.append(line)
            else:
                if '*/' in line:
                    part = line.split('*/')
                    ret.append(part[1])
                    incomment = False
                else:
                    pass
                
        lines = ret
        
        # remove \n
        ret = []
        for line in lines:
            parts = line.split('\n')
            for part in parts:
                if (len(part) > 0):
                    ret.append(part.strip()) 
                    
        return ret
    
    def load(self, filename):
        self.lines = self.readfile(filename)
        self.i =0
        
        self.top = VWFTop()
        self.parseObjects(self.top)
        
        #print('START DRAWING')
        #self.top.draw()
        
    
        
    def parseDict(self, obj):
        ret = {}
        
        line = self.lines[self.i].strip()
        self.i+=1
        
        if (self.verbose):
            print('{}:{}'.format(self.i, line))
        
        assert(line == '{')
        
        while (self.i < len(self.lines)):
            line = self.lines[self.i].strip()
            self.i+=1
            
            print('{}:{}'.format(self.i, line))
            
            if (line == '}'):
                obj.value = ret
                return 
            else:
                line = line[:-1] # trip final ;
                part = line.split('=')
                key = part[0].strip()
                value = part[1].strip()
                if (value[0] == '"'):
                    value = value[1:-1] # remove quotes
                    
                ret[key] = value
        
        assert(False)
           
    def parseList(self, obj):
        ret = []
        
        line = self.lines[self.i].strip()
        self.i+=1
        
        if (self.verbose):
            print('{}:{}'.format(self.i, line))
        
        assert(line == '{')
        
        while (self.i < len(self.lines)):
            line = self.lines[self.i].strip()
            self.i+=1
            
            print('{}:{}'.format(self.i, line))
            
            if (line == '}'):
                obj.value = ret
                return 
            else:
                if (startsWith(line, 'NODE')):
                    lobj = VWFNode()
                    self.parseList(lobj)
                    ret.append(lobj)                    
                else:
                    line = line[:-1] # trip final ;
                    ret.append(line)
        
        assert(False)
    
    def parseObjects(self, obj):
        if (self.verbose):
            print('parse', type(obj))

        while (self.i < len(self.lines)):
            line = self.lines[self.i].strip()
            self.i+=1
            
            if (self.verbose):
                print('{}:{}'.format(self.i, line))
            
            if (line == 'HEADER' ):
                lobj = VWFHeader()
                self.parseDict(lobj)
                obj.child.append(lobj)
            elif (line[0:7] == 'SIGNAL('):
                lobj = VWFSignal()
                lobj.parse(line)
                self.parseDict(lobj)
                obj.child.append(lobj)
            elif (line == 'TIME_BAR'):
                lobj = VWFTimeBar()
                self.parseDict(lobj)
                obj.child.append(lobj)
            elif (line == 'DISPLAY_LINE'):
                lobj = VWFDisplayLine()
                self.parseDict(lobj)
                obj.child.append(lobj)
            elif (line == 'NODE'):
                lobj = VWFNode()
                self.parseList(lobj)
                obj.child.append(lobj)
            elif (startsWith(line, 'TRANSITION_LIST(')):
                lobj = VWFTransitionList()
                lobj.parse(line)
                self.parseObjects(lobj)
                obj.child.append(lobj)
            elif (line == '{'):
                pass # ignore the start
            elif (line == '}'):
                #print('<<<<<', type(obj))
                return
            else:
                print('?? - ?? - ??', line)
                
    
    def draw_wavedrom(self):
        import nbwavedrom as wave
        return wave.draw(self.get_wavedrom())
    
    def get_wavedrom(self):
        
        header = self.top.find(VWFHeader)
        period = float(header.value['GRID_PERIOD'])
        
        #signals = [{"name": "clk", 'wave': 'P'}]
        signals = [] 
        
        for idx, tran in enumerate(self.top.findAll(VWFTransitionList)):
            node = tran.find(VWFNode)
            
            fmt = ''
            wavedata = 'x'
            wavedatadata = []
            
            data = convertTransitionsToValues(node.value, period)
            last = 'x'
            numclks = len(data)
            for i in range(numclks):
                v = data[i]
                if (v != last):
                    wavedata += '{}'.format(v)
                else:
                    wavedata += '.'
                last = v
                    
            wavedata += 'x'
            
            name = tran.name
                
            signals.append({'name': name, 'wave':wavedata, 'data':wavedatadata})

        #wavedata = 'P'
        #for i in range(numclks):
        #    wavedata += '.'
        #wavedata += 'x'

        #signals[0]['wave'] = wavedata
        
        ret = {
            "signal": signals,
            "head": {
                "text": 'Waveform',
                "tock": 0,
            }
        }
        
        return ret


# -*- coding: utf-8 -*-
"""
Created on Thu Oct 29 17:10:01 2020

@author: dcr
"""
from .. import *
from .bitwise import Buf

from deprecated import deprecated

class Latch(Logic):
    '''
    This is an asynchronous LATCH
    '''
    def __init__(self, parent, name, d:Wire, q:Wire, enable:Wire):
        super().__init__(parent, name)
        self.d = self.addIn("d", d)
        self.q = self.addOut("q", q)
        self.e = self.addIn("e", enable)
        
    def propagate(self):
        
        if (self.e.get()):
            self.q.put(self.d.get())
            
    def structureName(self):
        msg = 'Latch{}'.format(self.q.getWidth())
        return msg
    
class Reg(Logic):
    """
    This is a D flip flop with (optional) enable and (optional) reset
    """
    def __init__(self, parent, name:str, d:Wire, q:Wire, enable:Wire=None, reset:Wire=None, reset_value:int=None ):
        """
        
        D flip flop with (optional) enable and (optional) reset

        Parameters
        ----------
        parent : TYPE
            parent.
        name : str
            name.
        d : Wire
            input.
        q : Wire
            output.
        enable : Wire, optional
            Enable signal. The default is None.
        reset : Wire, optional
            Synchronous reset signal. The default is None.

        Returns
        -------
        the object.

        """
        super().__init__(parent, name)
        self.d = self.addIn("d", d)
        self.q = self.addOut("q", q)

        if not(enable is None):
            self.e = self.addIn("e", enable)
        else:
            self.e = None
            
        if not(reset is None):
            self.r = self.addIn("r", reset)
        else:
            self.r = None
            
        if not(reset_value is None):
            self.reset_value = reset_value
        else:
            self.reset_value = 0
            
        self.value = self.reset_value
        
    def clock(self):
        setValue = True
        resetValue = False

        if not(self.e is None):
            if (self.e.get() == 0):
                setValue = False
        if not(self.r is None):
            if (self.r.get() == 1):
                resetValue = 1
                
        if (resetValue):
            self.value = self.reset_value
        elif (setValue):
            self.value = self.d.get()
        #else:
        #   maintain the same value            
            
        self.q.prepare(self.value)

    def structureName(self):
        msg = 'Reg{}'.format(self.q.getWidth())
        
        if not(self.r is None): msg += 'R'
        if not(self.e is None): msg += 'E'
        if not(self.reset_value == 0): msg += '_v{}'.format(self.reset_value)
        
        return msg
            
class TReg(Logic):
    def __init__(self, parent, name:str, t:Wire, q:Wire, enable:Wire=None, reset:Wire=None ):
        from .bitwise import Not
        from .bitwise import Mux2
        
        super().__init__(parent, name)
        
        self.addIn("t", t)
        
        if not(enable is None):
            self.addIn("e", enable)

        if not(reset is None):
            self.addIn("r", reset)
        
        self.addOut("q", q)
        
        nq = self.wire('nq', 1)
        Not(self, 'nq', q, nq)
        
        d = self.wire('d', 1)
        Mux2(self, 'mux', t, q, nq, d)
        Reg(self, 'reg', d, q, enable=enable, reset=reset)
        
class DelayLine(Logic):
    def __init__(self, parent, name:str, a:Wire, en:Wire, reset:Wire, r:Wire, delay:int):
        """
        Daisy-chained number of registers, typically used to create a 
        fixed-length delay line
        
        Parameters
        ----------
        parent : Logic
            Parent entity.
        name : str
            instance name.
        a : Wire
            input.
        en : Wire
            enable signal.
        r : Wire
            output.
        delay : int
            number of registers in the delay line.

        Returns
        -------
        None.

        """
        
        super().__init__(parent, name)
        
        self.addIn('a', a)
        if not(en is None): self.addIn('en', en)
        if not(reset is None): self.addIn('reset', reset)
        self.addOut('r', r)

        last = a
        for i in range(delay):
            newlast = self.wire('r{}'.format(i), a.getWidth())
            Reg(self, 'r{}'.format(i), last, newlast, enable=en, reset=reset)
            last = newlast
            
        Buf(self, 'buf', last, r)
        
        
class PipelinePhase(Logic):
    def __init__(self, parent, name, reset, ins, outs):
        super().__init__(parent, name)
        assert(len(ins) == len(outs))
        
        self.addIn('reset', reset)
        
        for i in range(len(ins)):
            self.addIn('in{}'.format(i), ins[i])
            self.addOut('out{}'.format(i), outs[i])
            
            Reg(self, 'r{}'.format(i), ins[i], outs[i], reset=reset)
            
            
class AsynchronousMemory(Logic):
    def __init__(self, parent, name, read_address, write_address, write, readdata, writedata):
        super().__init__(parent, name)
        
        self.read_address = self.addIn('read_address', read_address)
        self.write_address = self.addIn('write_address', write_address)
        self.write = self.addIn('write', write)
        self.readdata = self.addOut('readdata', readdata)
        self.writedata = self.addIn('writedata', writedata)
        
        if (read_address.getWidth() != write_address.getWidth()):
            raise Exception(f'read and write address must have the same width')
            
        if (read_address.getWidth() > 10):
            raise Exception(f'Memory too big! address width = {address.getWidth()}')
            
        numcells = 1 << read_address.getWidth()
        
        
        self.data = [0] * numcells
        
    def propagate(self):
        radd = self.read_address.get()
        wadd = self.write_address.get()
        
        # always reading
        #print(f'reading address {add} = {self.data[add]}')
        self.readdata.put(self.data[radd])
        
        if (self.write.get()):
            self.data[wadd] = self.writedata.get()
            
        
    def verilogBody(self):
        
        numcells = 1 << self.read_address.getWidth()
        w = self.readdata.getWidth()
        
        s = f'reg [{w-1}:0] mem [{numcells-1}:0];\n'
        s += "assign readdata =  mem[read_address];\n"

        s += 'always @(*) begin\n'
        s += 'if (write) \n'
        s += ' mem[write_address] = writedata;\n'
        s += 'end\n'
    
        return s
        
    
    


class SynchronousMemory(Logic):
    def __init__(self, parent, name, read_address, write_address, write, readdata, writedata):
        super().__init__(parent, name)
        
        self.read_address = self.addIn('read_address', read_address)
        self.write_address = self.addIn('write_address', write_address)
        self.write = self.addIn('write', write)
        self.readdata = self.addOut('readdata', readdata)
        self.writedata = self.addIn('writedata', writedata)
        
        if (read_address.getWidth() != write_address.getWidth()):
            raise Exception(f'read and write address must have the same width')
            
        if (read_address.getWidth() > 10):
            raise Exception(f'Memory too big! address width = {address.getWidth()}')
            
        numcells = 1 << read_address.getWidth()
        
        
        self.data = [0] * numcells
        
    def clock(self):
        radd = self.read_address.get()
        wadd = self.write_address.get()
        
        # always reading
        #print(f'reading address {add} = {self.data[add]}')
        self.readdata.prepare(self.data[radd])
        
        if (self.write.get()):
            self.data[wadd] = self.writedata.get()
            


    def verilogBody(self):
        numcells = 1 << self.read_address.getWidth()
        w = self.readdata.getWidth()
        
        s = f'(* ramstyle = "no_rw_check" *) reg [{w-1}:0] mem [0:{numcells-1}];\n'

        s += f'reg [{w-1}:0] rreaddata;\n'
        s += 'always @(posedge clk) begin\n'
        s += 'if (write) \n'
        s += ' mem[write_address] <= writedata;\n'
        s += ' rreaddata <= mem[read_address];\n'
        s += 'end\n'
    
        s += 'assign readdata = rreaddata;\n'

        return s
        
        
class DualPortSynchronousMemory(Logic):
    def __init__(self, parent, name, read_address_a, write_address_a, write_a, readdata_a, writedata_a,
                 read_address_b, write_address_b, write_b, readdata_b, writedata_b):
        super().__init__(parent, name)
        
        self.read_address_a = self.addIn('read_address_a', read_address_a)
        self.write_address_a = self.addIn('write_address_a', write_address_a)
        self.write_a = self.addIn('write_a', write_a)
        self.readdata_a = self.addOut('readdata_a', readdata_a)
        self.writedata_a = self.addIn('writedata_a', writedata_a)

        self.read_address_b = self.addIn('read_address_b', read_address_b)
        self.write_address_b = self.addIn('write_address_b', write_address_b)
        self.write_b = self.addIn('write_b', write_b)
        self.readdata_b = self.addOut('readdata_b', readdata_b)
        self.writedata_b = self.addIn('writedata_b', writedata_b)
        
        if ((read_address_a.getWidth() != write_address_a.getWidth()) or
            (read_address_b.getWidth() != write_address_b.getWidth()) or
            (read_address_a.getWidth() != read_address_b.getWidth())):
            raise Exception(f'read and write address must have the same width')
            
        if (read_address_a.getWidth() > 10):
            raise Exception(f'Memory too big! address width = {address.getWidth()}')
            
        numcells = 1 << read_address_a.getWidth()
               
        self.data = [0] * numcells
        
    def clock(self):
        radda = self.read_address_a.get()
        wadda = self.write_address_a.get()

        raddb = self.read_address_b.get()
        waddb = self.write_address_b.get()
        
        # always reading
        #print(f'reading address {add} = {self.data[add]}')
        self.readdata_a.prepare(self.data[radda])
        
        if (self.writea.get()):
            self.data[wadd] = self.writedataa.get()
            
        self.readdata_b.prepare(self.data[raddb])
        
        if (self.writeb.get()):
            self.data[wadd] = self.writedatab.get()


    def verilogBody(self):
        numcells = 1 << self.read_address_a.getWidth()
        w = self.readdata_a.getWidth()
        
        s = f'reg [{w-1}:0] mem [{numcells-1}:0];\n'

        s += 'always @(posedge clk) begin\n'
        s += 'if (write_a) \n'
        s += ' mem[write_address_a] <= writedata_a;\n'
        s += 'end\n'

        s += 'always @(posedge clk) begin\n'
        s += 'if (write_b) \n'
        s += ' mem[write_address_b] <= writedata_b;\n'
        s += 'end\n'

    
        s += 'assign readdata_a = mem[read_address_a];\n'
        s += 'assign readdata_b = mem[read_address_b];\n'
        return s
        

class ShiftRegisterBidirectional(Logic):
    def __init__(self, parent, name, 
                 left_in:Wire, right_in:Wire, 
                 left_out:Wire, right_out:Wire, 
                 shift_left:Wire, shift_right:Wire, depth:int):
        
        from .bitwise import Or2
        from .bitwise import Mux2
        
        super().__init__(parent, name)
        
        w = left_in.getWidth()
        assert(w == left_out.getWidth())
        assert(w == right_in.getWidth())
        assert(w == right_out.getWidth())
                
        self.addIn('left_in', left_in)
        self.addIn('right_in', right_in)
        self.addOut('left_out', left_out)
        self.addOut('right_out', right_out)
        
        self.addIn('shift_left', shift_left)
        self.addIn('shift_right', shift_right)
        
        q = self.wires('q', depth, w)
        
        # dir_r wires: left_in -> [0] -> [1] -> [2] -> right_out 
        # dir_l wires: left_out <- [0] <- [1] <- [2] <- right_in
        shift = self.wire('shift')
        Or2(self, 'shift', shift_left, shift_right, shift)
        
        for i in range(depth):
            if (i == 0):
                vl = left_in
            else:
                vl = q[i-1]
                
            if (i == depth-1):
                vr = right_in
            else:
                vr = q[i+1]
            
            rd = self.wire(f'rd{i}', w)
            # if we are shifting left, the value to input in the register 
            # should be the right one
            Mux2(self, f'rd{i}', shift_left, vl, vr, rd )
            Reg(self, f'r{i}', d=rd, q=q[i], enable=shift)
        
        Buf(self, 'left_out', q[0], left_out)
        Buf(self, 'right_out', q[depth-1], right_out)
        
class Stack_ShiftRegister(Logic):
    def __init__(self, parent, name, din, dout, push, pop, empty, full, depth):
        
        from .bitwise import Constant
        
        super().__init__(parent, name)


        # IO
        self.addIn("din", din)
        self.addIn("push", push)
        self.addIn("pop", pop)
        self.addOut("dout", dout)
        
        if not(empty is None):
            self.addOut("empty", empty)
            
        if not(full is None):
            self.addOut("full", full)

        zerow = self.wire('zerow', din.getWidth())
        pre_dout = self.wire('pre_dout', dout.getWidth())
        rout = self.wire('rout', dout.getWidth())
        
        Constant(self, 'zerow', 0 , zerow)
        ShiftRegisterBidirectional(self, 'shift', din, zerow, pre_dout, rout, pop, push, depth)
        
        # poped value is always present at the left side of the shift register, just load it 
        # into the register when poping
        Reg(self, 'dout', pre_dout, dout, enable=pop)
        
        



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 13 06:51:58 2025

@author: 2016570
"""
import ast
import textwrap
import inspect
import copy
from typing import Dict, Any

class YieldFromInliner(ast.NodeTransformer):
    def __init__(self, cls_node):
        self.cls_node = cls_node
        self.inlinable_methods = {}
        self.calls_to_inline = [] # This is primarily for debugging/information in Phase 1
        self.methods_to_remove = set()
        self._phase1_collect_info()
        self.current_function_stack = []

    def _phase1_collect_info(self):
        # print("\n--- Phase 1: Collecting Information ---")
        for node in self.cls_node.body:
            if isinstance(node, ast.FunctionDef):
                self.inlinable_methods[node.name] = node
                # print(f'  Collecting method: {node.name}')

        call_collector = CallSiteCollector(self.inlinable_methods.keys())
        call_collector.visit(self.cls_node)
        self.calls_to_inline = call_collector.collected_calls
        # print(f'  Identified {len(self.calls_to_inline)} potential inlining candidates.')

    def _phase2_apply_inlining(self):
        # print("\n--- Phase 2: Applying Inlining and Removing Functions ---")
        transformed_cls_node = self.visit(self.cls_node) # This kicks off the transformation

        new_class_body = []
        removed_count = 0
        for node in transformed_cls_node.body:
            if isinstance(node, ast.FunctionDef) and node.name in self.methods_to_remove:
                # print(f"  Removing inlined function definition: {node.name}")
                removed_count += 1
                continue
            new_class_body.append(node)
        transformed_cls_node.body = new_class_body
        # print(f"  Removed {removed_count} function definitions.")

        return transformed_cls_node

    def visit_FunctionDef(self, node):
        self.current_function_stack.append(node.name)
        # It's important to process the body recursively here to ensure nested inlines
        # The generic_visit calls visit for each child.
        # However, for function body, we need to manage it to ensure newly inserted
        # statements are also processed.
        
        # We need to manually iterate and visit statements in the body,
        # handling lists returned by visit for inlining.
        new_body = []
        for stmt in node.body:
            transformed_stmt = self.visit(stmt) # Visit each statement
            if isinstance(transformed_stmt, list):
                # If a statement was inlined into multiple statements, extend the list
                new_body.extend(transformed_stmt)
            elif transformed_stmt is not None: # Don't add if node was removed (returned None)
                new_body.append(transformed_stmt)
        node.body = new_body
        
        self.current_function_stack.pop()
        return node

    def visit_Expr(self, node):
        # First, ensure nested expressions are visited
        node = self.generic_visit(node)

        if isinstance(node.value, ast.YieldFrom):
            call = node.value.value
            if isinstance(call, ast.Call) and isinstance(call.func, ast.Attribute):
                if isinstance(call.func.value, ast.Name) and call.func.value.id == 'self':
                    method_name = call.func.attr
                    if method_name in self.inlinable_methods:
                        if self.current_function_stack and method_name == self.current_function_stack[-1]:
                            raise RecursionError(f"Refusing to inline recursive call to {method_name}")

                        method_node = self.inlinable_methods[method_name]
                        
                        # Deep copy the body of the method being inlined.
                        # Then, RECURSIVELY VISIT EACH STATEMENT IN THE INLINED BODY.
                        # This is the crucial change for nested inlining.
                        transformed_inlined_body = []
                        for stmt in method_node.body:
                            # Visit the statement to handle any further inlining *within* it
                            visited_stmt = self.visit(copy.deepcopy(stmt))
                            if isinstance(visited_stmt, list):
                                transformed_inlined_body.extend(visited_stmt)
                            elif visited_stmt is not None:
                                transformed_inlined_body.append(visited_stmt)

                        # print(f"  Inlining '{method_name}' into '{self.current_function_stack[-1]}'")
                        self.methods_to_remove.add(method_name)
                        return transformed_inlined_body
        return node

    def visit_Return(self, node):
        # This is a placeholder. Real inlining would need to convert
        # 'return value' from an inlined generator into the final value
        # of the 'yield from' expression, which is complex.
        # For simple inlining, you might remove it or transform to 'pass'.
        return self.generic_visit(node)



class YieldStateTransformer(ast.NodeTransformer):
    def __init__(self):
        self.yield_counter = 1

    def visit_ClassDef(self, node):
        # Visit all methods first
        self.generic_visit(node)

        # Check if __init__ method exists
        init_found = False
        for stmt in node.body:
            if isinstance(stmt, ast.FunctionDef) and stmt.name == "__init__":
                init_found = True
                # Insert self.state = 0 at the beginning of __init__
                assign_stmt = ast.parse("self.state = 0").body[0]
                stmt.body.insert(0, assign_stmt)
                break

        # If __init__ not found, create one with self.state = 0
        if not init_found:
            init_func = ast.FunctionDef(
                name="__init__",
                args=ast.arguments(
                    posonlyargs=[],
                    args=[ast.arg(arg='self')],
                    vararg=None, kwonlyargs=[], kw_defaults=[],
                    kwarg=None, defaults=[]
                ),
                body=[ast.parse("self.state = 0").body[0]],
                decorator_list=[]
            )
            node.body.insert(0, init_func)

        return node

    def visit_FunctionDef(self, node):
        # Only transform the "run" method
        if node.name == "run":
            self.generic_visit(node)
        return node

    def visit_Yield(self, node):
        # Replace each yield with: self.state = N
        state_assign = ast.Assign(
            targets=[ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr='state', ctx=ast.Store())],
            value=ast.Constant(value=self.yield_counter)
        )
        self.yield_counter += 1
        return state_assign




class CallSiteCollector(ast.NodeVisitor):
    # (No changes needed for CallSiteCollector for this particular fix)
    def __init__(self, inlinable_method_names):
        self.inlinable_method_names = inlinable_method_names
        self.collected_calls = []
        self.current_function_stack = []

    def visit_FunctionDef(self, node):
        self.current_function_stack.append(node.name)
        self.generic_visit(node)
        self.current_function_stack.pop()

    def visit_YieldFrom(self, node):
        call = node.value
        if isinstance(call, ast.Call) and isinstance(call.func, ast.Attribute):
            if isinstance(call.func.value, ast.Name) and call.func.value.id == 'self':
                method_name = call.func.attr
                if method_name in self.inlinable_method_names:
                    if self.current_function_stack:
                        calling_function = self.current_function_stack[-1]
                        self.collected_calls.append((calling_function, method_name, call))
        self.generic_visit(node)


class KeepInitAndClockTransformer(ast.NodeTransformer):
    def visit_ClassDef(self, node):
        # Keep only methods named '__init__' or 'clock'
        node.body = [
            item for item in node.body
            if not isinstance(item, ast.FunctionDef)
            or item.name in ('__init__', 'clock')
        ]
        self.generic_visit(node)
        return node
    
class IOWireDirectoryDetector(ast.NodeVisitor):
    def __init__(self):
        self.input_dir = None
        self.output_dir = None

    def visit_FunctionDef(self, node):
        if node.name != '__init__':
            return  # Only inspect __init__
        for stmt in node.body:
            if isinstance(stmt, ast.For):
                self._process_for_loop(stmt)

    def _process_for_loop(self, node):
        # Check if loop iterates over something like: <dict>.keys()
        if (
            isinstance(node.iter, ast.Call) and
            isinstance(node.iter.func, ast.Attribute) and
            node.iter.func.attr == 'keys' and
            isinstance(node.iter.func.value, ast.Name)
        ):
            dict_var = node.iter.func.value.id  # the dict being iterated over
            for inner_stmt in node.body:
                # Check for: self.addIn(...) or self.addOut(...)
                if isinstance(inner_stmt, ast.Expr):
                    call = inner_stmt.value
                    if (
                        isinstance(call, ast.Call) and
                        isinstance(call.func, ast.Attribute) and
                        isinstance(call.func.value, ast.Name) and
                        call.func.value.id == 'self'
                    ):
                        method = call.func.attr
                        if method == 'addIn' and self.input_dir is None:
                            self.input_dir = dict_var
                        elif method == 'addOut' and self.output_dir is None:
                            self.output_dir = dict_var


class ReplacePrintByPass(ast.NodeTransformer):
        
    def visit_Call(self, node):
        from py4hw.rtl_generation import getAstName

        attr = getAstName(node.func)
        
        #print('checking call', attr)
        if (attr == 'print'):
            # remove prints
            return ast.Pass()
        
        node = ast.NodeTransformer.generic_visit(self, node)
        
        return node


import ast
from typing import Any, Dict, Tuple

# ---------- Helper ----------
def dotted_name_from_attribute(node: ast.Attribute) -> str | None:
    """Return dotted name like 'ALU.OP_INC' or 'pkg.mod.Class.CONST'."""
    parts = []
    cur = node
    while isinstance(cur, ast.Attribute):
        parts.append(cur.attr)
        cur = cur.value
    if isinstance(cur, ast.Name):
        parts.append(cur.id)
        return '.'.join(reversed(parts))
    return None

class ClassConstantCollector(ast.NodeVisitor):
    def __init__(self, namespace):
        self.constants = {}
        self.current_class = None
        self.namespace = namespace

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        """Collect class-level constants (case 1)"""
        self.current_class = node.name
        for item in node.body:
            if isinstance(item, ast.Assign):
                for target in item.targets:
                    if isinstance(target, ast.Name):
                        try:
                            value = ast.literal_eval(item.value)
                            self.constants[f"self.{target.id}"] = value
                        except (ValueError, SyntaxError):
                            pass
        self.generic_visit(node)  # Continue visiting child nodes

    def visit_Attribute(self, node: ast.Attribute) -> None:
        """Collect external class constants (case 2)"""
        if isinstance(node.value, ast.Name):
            # Pattern: OtherClass.CONST
            print('Attribute', f"{node.value.id}.{node.attr}")
            try:
                value = eval(f"{node.value.id}.{node.attr}", self.namespace)
                self.constants[f"{node.value.id}.{node.attr}"] = value
            except :
                print('error')
                pass
        elif (isinstance(node.value, ast.Attribute) and 
              isinstance(node.value.value, ast.Name) and 
              node.value.value.id == 'self'):
            # Pattern: self.CONST (already handled in visit_ClassDef)
            pass
        self.generic_visit(node)
        
# ---------- Transformer with eval fallback ----------
class StaticAttributeInliner(ast.NodeTransformer):
    def __init__(
        self,
        ast_constants: Dict[str, Any],
        runtime_namespace: Dict[str, Any],
        allowed_types: Tuple[type, ...] = (int, float, str, bool, tuple)
    ):
        self.constants = dict(ast_constants)
        self.runtime_namespace = runtime_namespace
        self.allowed_types = allowed_types
        self._current_class = None
        self._in_bases = False  # Flag to track if visiting bases

    def visit_ClassDef(self, node: ast.ClassDef):
        prev_class = self._current_class
        self._current_class = node.name

        # First visit bases with flag set
        self._in_bases = True
        for i, base in enumerate(node.bases):
            node.bases[i] = self.visit(base)
        self._in_bases = False

        # Then visit the rest of the body normally
        node.body = [self.visit(stmt) for stmt in node.body]

        self._current_class = prev_class
        return node

    def visit_Attribute(self, node: ast.Attribute):
        # When inside base classes, do not replace anything
        if self._in_bases:
            return self.generic_visit(node)

        self.generic_visit(node)

        # Case 1: self.CONST inside current class
        if isinstance(node.value, ast.Name) and node.value.id == "self" and self._current_class:
            key = f"{self._current_class}.{node.attr}"
            if key in self.constants:
                return ast.copy_location(ast.Constant(value=self.constants[key]), node)

        # Case 2: Class.CONST or module.Class.CONST
        dotted = dotted_name_from_attribute(node)
        if dotted:
            if dotted in self.constants:
                return ast.copy_location(ast.Constant(value=self.constants[dotted]), node)

            parts = dotted.split('.')
            if len(parts) >= 2:
                last_two = '.'.join(parts[-2:])
                if last_two in self.constants:
                    return ast.copy_location(ast.Constant(value=self.constants[last_two]), node)

            try:
                val = eval(dotted, self.runtime_namespace)
                if isinstance(val, self.allowed_types):
                    self.constants[dotted] = val
                    if len(parts) >= 2:
                        self.constants[last_two] = val
                    return ast.copy_location(ast.Constant(value=val), node)
            except Exception:
                pass

        return node


def replaceConstants(tree:ast.AST, clazz):
    # Execute source to create namespace
    module = inspect.getmodule(clazz)
    src = inspect.getsource(module)
    namespace = {}
    
    print(src[0:1000])
    exec(src, namespace)
    
    collector = ClassConstantCollector(namespace)
    collector.visit(tree)

    print('Collected constants', collector.constants)
    transformer = StaticAttributeInliner(collector.constants, globals())
    new_tree = transformer.visit(tree)

    return new_tree

class ScopedSelfAliasRewriter(ast.NodeTransformer):
    def __init__(self):
        self.current_aliases = {}

    def visit_FunctionDef(self, node):
        saved_aliases = self.current_aliases
        self.current_aliases = {}

        node = self.generic_visit(node)

        # Remove alias assignments like a = self.<something>
        new_body = []
        for stmt in node.body:
            if (isinstance(stmt, ast.Assign) and
                len(stmt.targets) == 1 and
                isinstance(stmt.targets[0], ast.Name) and
                self._is_self_expr(stmt.value)):
                continue  # skip alias
            new_body.append(stmt)

        self.current_aliases = saved_aliases
        node.body = new_body
        return node

    def visit_Assign(self, node):
        if (
            len(node.targets) == 1 and
            isinstance(node.targets[0], ast.Name) and
            self._is_self_expr(node.value)
        ):
            alias_name = node.targets[0].id
            self.current_aliases[alias_name] = copy.deepcopy(node.value)
            return None  # remove assignment
        return self.generic_visit(node)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load) and node.id in self.current_aliases:
            return copy.deepcopy(self.current_aliases[node.id])
        return node

    def _is_self_expr(self, expr):
        """
        Returns True if expr starts from `self`, e.g., self.foo.bar()['x'].baz
        """
        while isinstance(expr, (ast.Attribute, ast.Subscript, ast.Call)):
            if isinstance(expr, ast.Call):
                expr = expr.func
            elif isinstance(expr, ast.Subscript):
                expr = expr.value
            else:  # ast.Attribute
                expr = expr.value
        return isinstance(expr, ast.Name) and expr.id == 'self'



class IODictKeyCollector(ast.NodeVisitor):
    def __init__(self, input_dir_name, output_dir_name, method_name):
        self.input_dir_name = input_dir_name
        self.output_dir_name = output_dir_name
        self.method_name = method_name
        self.input_keys = set()
        self.output_keys = set()

    def visit_FunctionDef(self, node):
        print('checking function ', node.name ,  self.method_name)
        if node.name != self.method_name:
            return
        self.generic_visit(node)

    def visit_Subscript(self, node):
        # Check if we're subscript-ing an attribute like self.status["key"]
        import astunparse
        #print('Checking subscript', astunparse.unparse(node), type(node.value))
        target = node.value
        if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name):
            if target.value.id == 'self':
                dict_name = target.attr  # e.g., 'status' or 'control'

                # Handle dict["key"] where key is a constant string
                if isinstance(node.slice, ast.Constant):  # Python 3.9+
                    key = node.slice.value
                elif isinstance(node.slice, ast.Index) and isinstance(node.slice.value, ast.Constant):  # pre-3.9
                    key = node.slice.value.value
                else:
                    key = None

                if isinstance(key, str):
                    if dict_name == self.input_dir_name:
                        self.input_keys.add(key)
                    elif dict_name == self.output_dir_name:
                        self.output_keys.add(key)
        self.generic_visit(node)





import itertools

class SSARewriter(ast.NodeTransformer):
    def __init__(self):
        self.counter = itertools.count()
        self.versions = {}  # variable → latest version name
        self.rename_map = {}  # current loads → latest version name

    def visit_Assign(self, node):
        # Only handle simple single-target assignments
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            orig_name = node.targets[0].id

            # Visit RHS first, so it uses the *old* version
            node.value = self.visit(node.value)

            # Then create new version for LHS
            new_name = f"{orig_name}_{next(self.counter)}"
            self.versions[orig_name] = new_name
            self.rename_map[orig_name] = new_name
            node.targets[0].id = new_name
            return node

        return self.generic_visit(node)

    def visit_Name(self, node):
        # For loads, rewrite to latest version
        if isinstance(node.ctx, ast.Load) and node.id in self.rename_map:
            node.id = self.rename_map[node.id]
        return node


class RemoveUnusedAssignments(ast.NodeTransformer):
    def __init__(self):
        self.changed = False
        
    def visit_FunctionDef(self, node):
        self.generic_visit(node)  # process children first

        assigned_vars = set()
        used_vars = set()

        # Collect assignments and uses throughout the entire function
        for stmt in ast.walk(node):
            if isinstance(stmt, ast.Assign):
                for target in stmt.targets:
                    for t in ast.walk(target):
                        if isinstance(t, ast.Name) and isinstance(t.ctx, ast.Store):
                            assigned_vars.add(t.id)
            elif isinstance(stmt, ast.AnnAssign):
                target = stmt.target
                if isinstance(target, ast.Name) and isinstance(target.ctx, ast.Store):
                    assigned_vars.add(target.id)
            elif isinstance(stmt, ast.Name) and isinstance(stmt.ctx, ast.Load):
                used_vars.add(stmt.id)

        unused_vars = assigned_vars - used_vars
        
        if (len(unused_vars) > 0):
            self.changed = True
            print('ununsed', unused_vars)

        # Now filter out assignments to only-unused variables
        def is_unused_assignment(stmt):
            def extract_names(target):
                if isinstance(target, ast.Name):
                    return [target.id]
                elif isinstance(target, ast.Tuple):
                    return [id for elt in target.elts for id in extract_names(elt)]
                return []
        
            if isinstance(stmt, ast.Assign):
                all_names = []
                for t in stmt.targets:
                    all_names.extend(extract_names(t))
                return all_names and all(name in unused_vars for name in all_names)
        
            elif isinstance(stmt, ast.AnnAssign):
                if isinstance(stmt.target, ast.Name):
                    return stmt.target.id in unused_vars
        
            return False


        class AssignmentCleaner(ast.NodeTransformer):
            def visit_Assign(self, stmt):
                return ast.Pass() if is_unused_assignment(stmt) else stmt

            def visit_AnnAssign(self, stmt):
                return ast.Pass() if is_unused_assignment(stmt) else stmt

        node.body = [AssignmentCleaner().visit(stmt) for stmt in node.body]
        return node


class RemoveIrrelevantConditions(ast.NodeTransformer):
    def visit_If(self, node):
        self.generic_visit(node)  # first transform inside the body

        ret = False
        
        if len(node.body) == 0:
            ret = True
        elif len(node.body) == 1:
            if (isinstance(node.body[0], ast.Pass)):
                ret = True
            elif (isinstance(node.body[0], ast.Expr)) and (isinstance(node.body[0].value, ast.Pass)):
                ret = True
        
        if len(node.orelse) != 0:
            # how could this ever change ?
            ret = False

        if ret :
            return None  

        return node


class ExpandInitLoopTransformer(ast.NodeTransformer):
    def __init__(self, input_dir, output_dir, input_keys, output_keys):
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.input_keys = input_keys
        self.output_keys = output_keys
        self.all_keys = input_keys.union(output_keys)

    def visit_FunctionDef(self, node):
        if node.name == '__init__':
            # Rewrite constructor loops
            new_body = []
            for stmt in node.body:
                if self._is_status_loop(stmt):
                    new_body.extend(self._generate_input_assignments())
                elif self._is_control_loop(stmt):
                    new_body.extend(self._generate_output_assignments())
                else:
                    new_body.append(stmt)
            node.body = new_body
        else:
            # Rewrite Subscript accesses elsewhere (like in clock)
            self.generic_visit(node)
        return node

    def visit_Subscript(self, node):
        # Only replace self.status["key"] or self.control["key"]
        if (
            isinstance(node.value, ast.Attribute) and
            isinstance(node.value.value, ast.Name) and
            node.value.value.id == 'self' and
            node.value.attr in {self.input_dir, self.output_dir}
        ):
            key = None
            if isinstance(node.slice, ast.Constant):  # Python 3.9+
                key = node.slice.value
            elif isinstance(node.slice, ast.Index) and isinstance(node.slice.value, ast.Constant):  # <3.9
                key = node.slice.value.value
            if isinstance(key, str) and key in self.all_keys:
                return ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr=key, ctx=node.ctx)
        return self.generic_visit(node)

    def _is_status_loop(self, node):
        return (
            isinstance(node, ast.For) and
            isinstance(node.iter, ast.Call) and
            isinstance(node.iter.func, ast.Attribute) and
            node.iter.func.attr == 'keys' and
            isinstance(node.iter.func.value, ast.Name) and
            node.iter.func.value.id == self.input_dir
        )

    def _is_control_loop(self, node):
        return (
            isinstance(node, ast.For) and
            isinstance(node.iter, ast.Call) and
            isinstance(node.iter.func, ast.Attribute) and
            node.iter.func.attr == 'keys' and
            isinstance(node.iter.func.value, ast.Name) and
            node.iter.func.value.id == self.output_dir
        )

    def _generate_input_assignments(self):
        return [
            ast.Assign(
                targets=[ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr=key, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr='addIn', ctx=ast.Load()),
                    args=[
                        ast.Constant(value=key),
                        ast.Subscript(
                            value=ast.Name(id=self.input_dir, ctx=ast.Load()),
                            slice=ast.Constant(value=key),
                            ctx=ast.Load()
                        )
                    ],
                    keywords=[]
                )
            )
            for key in sorted(self.input_keys)
        ]

    def _generate_output_assignments(self):
        return [
            ast.Assign(
                targets=[ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr=key, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), attr='addOut', ctx=ast.Load()),
                    args=[
                        ast.Constant(value=key),
                        ast.Subscript(
                            value=ast.Name(id=self.output_dir, ctx=ast.Load()),
                            slice=ast.Constant(value=key),
                            ctx=ast.Load()
                        )
                    ],
                    keywords=[]
                )
            )
            for key in sorted(self.output_keys)
        ]



def getClassTree(cls: type) -> ast.AST:
    """
    Transforms a generator-based class's 'run' method into a state machine 'clock' method.

    Args:
        cls: The class object to transform.

    Returns:
        The AST of the class
    """
    if not isinstance(cls, type):
        raise TypeError("Expected a class object")

    try:
        source = inspect.getsource(cls)
    except Exception as e:
        raise ValueError(f"Could not retrieve source for {cls.__name__}: {e}")

    source = textwrap.dedent(source)
    tree = ast.parse(source)
    
    class_node = next((n for n in tree.body if isinstance(n, ast.ClassDef) and n.name == cls.__name__), None)
    if class_node is None:
        raise ValueError(f"Class {cls.__name__} not found in AST.")


    return class_node

# --- Helper function to apply the transformer ---
def transform_generator_to_fsm(class_node: ast.AST) -> ast.AST:

    #print(astunparse.unparse(class_node))

    transformer = GeneratorToFSMTransformer()
    tree = transformer.visit(class_node)

    #for i, node in enumerate(tree.body):
    #    if isinstance(node, ast.ClassDef) and node.name == cls.__name__:
    #        tree.body[i] = transformed_class_node
    #        break

    ast.fix_missing_locations(tree)
    
    return tree
    #return astunparse.unparse(tree)


def removeUnusedAssignments(tree: ast.AST) -> ast.AST:
    tr = RemoveUnusedAssignments()
    doRun = True
    while doRun:
        tr.changed = False 
        tree = tr.visit(tree)
        doRun = tr.changed

    return tree

def inline_yield_from(class_node: ast.AST) -> ast.AST:
   

    # Phase 1: Initialize the transformer, which collects information
    transformer = YieldFromInliner(class_node)

    # Phase 2: Apply the transformations and remove the inlined functions
    # The _phase2_apply_inlining method itself performs the visit operation internally
    tree = transformer._phase2_apply_inlining()

    # Replace the original class_node in the tree with the transformed one
    #for i, node in enumerate(tree.body):
    #    if isinstance(node, ast.ClassDef) and node.name == cls.__name__:
    #        tree.body[i] = transformed_class_node
    #        break

    # Fix line numbers and other metadata in the AST
    ast.fix_missing_locations(tree)

    # Convert back to code
    #return astunparse.unparse(tree)
    return tree

def yield_to_state(class_node: ast.AST) -> ast.AST:
    tr = YieldStateTransformer()
    tree = tr.visit(class_node)
    ast.fix_missing_locations(tree)
    return tree



class ReplaceIfElseTreeByMatchCase(ast.NodeTransformer):
    """Transforms if/elif chains on the same variable into a match/case AST."""

    def visit_If(self, node):
        match_node = self.try_match_if_chain(node)
        if match_node:
            return match_node
        return self.generic_visit(node)

    def try_match_if_chain(self, node):
        var_expr = None
        case_list = []
        default_body = None
        cur = node

        while True:
            # Must be a Compare node
            if not isinstance(cur.test, ast.Compare):
                return None
            cmp = cur.test

            # Check '==' and single comparator
            if len(cmp.ops) != 1 or not isinstance(cmp.ops[0], ast.Eq):
                return None
            if len(cmp.comparators) != 1:
                return None

            lhs = cmp.left
            rhs = cmp.comparators[0]

            # First branch: store the variable expression
            if var_expr is None:
                var_expr = lhs
            elif ast.dump(var_expr) != ast.dump(lhs):
                return None  # expression differs

            # Create a MatchValue pattern
            pattern = ast.MatchValue(value=rhs)
            case_list.append(ast.match_case(pattern=pattern, guard=None, body=cur.body))

            # Go to elif
            if len(cur.orelse) == 1 and isinstance(cur.orelse[0], ast.If):
                cur = cur.orelse[0]
            else:
                if cur.orelse:
                    default_body = cur.orelse
                break

        if len(case_list) >= 2:
            if default_body:
                case_list.append(
                    ast.match_case(
                        pattern=ast.MatchAs(name=None),  # case _
                        guard=None,
                        body=default_body
                    )
                )
            return ast.Match(subject=var_expr, cases=case_list)

        return None



# -*- coding: utf-8 -*-
"""
Created on Sat Dec 21 13:17:38 2024

@author: 2016570
"""

import py4hw
from edalize import edatool
import os

class C10LP(py4hw.HWSystem):
    
    def __init__(self):
        
        super().__init__(name='C10LP')
        
        clk50 = self.wire('c10_clk50m')
        #clk50 = self.addIn('CLOCK_50', clk50)
        clockDriver = py4hw.ClockDriver('c10_clk50m', 50E6, 0, wire=clk50)
        
        print('Initial', self.getFullPath(),  self._wires)
        
        self.clockDriver = clockDriver
    
    
    
    def build(self,  projectDir, createdStructures=[]):
        if not(os.path.exists(projectDir)):
            print('Creating the directory', projectDir)
            os.makedirs(projectDir)
            
        rtl = py4hw.VerilogGenerator(self)
        
        rtl_code = rtl.getVerilogForHierarchy(noInstanceNumberInTopEntity=True, createdStructures=createdStructures)
        qsf_code = self.getQsf()
        top_name = self.name
        
        verilog_file = os.path.join(projectDir, top_name+'.v')
        qsf_file = os.path.join(projectDir, top_name+'_pins.qsf')
        sdc_file = os.path.join(projectDir, 'CLOCK_50.sdc')

        # Create Verilog file
        with open(verilog_file, 'w') as file:
            file.write(rtl_code)

        # Create Verilog file
        with open(qsf_file, 'w') as file:
            file.write(qsf_code)
            
        with open(sdc_file, 'w') as file:
            file.write('create_clock  -period 20 [get_ports c10_clk50m]\n')
            
        tool = 'quartus'
        
        files = [  {'name' : verilog_file, 'file_type' : 'verilogSource'},
                 {'name' : sdc_file, 'file_type' : 'SDC'},
                 {'name' : qsf_file, 'file_type' : 'tclSource'}
                 ]

        info = self.getFPGADeviceInfo()

        parameters = {'clk_freq_hz' : {'datatype' : 'int', 'default' : 50000000, 'paramtype' : 'vlogparam'},
              'vcd' : {'datatype' : 'bool', 'paramtype' : 'plusarg'},
              'family': {'datatype': 'string', 'paramtype': 'generic', 'default':'Cyclone 10'},
              'device': {'datatype': 'string', 'paramtype': 'generic', 'default':info['device']},
              }

        tool_options = {'family':info['family'],
                        'flags': ['--64bit'],
                        'quartus': {'flags': ['--64bit']}
                        }

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
          'tool_options' : tool_options
        }        
        
        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)
        
        print('Configure Edalize')
        backend.configure()
        
        print('Build Edalize')
        backend.build()
    
    def download(self, projectDir):
        tool='quartus'
        top_name = self.name
        
        files=[]
        parameters={
              'cable': {'datatype': 'string', 'paramtype': 'generic', 'default':'USB-Blaster'},
              'board_device_index': {'datatype': 'string', 'paramtype': 'generic', 'default':'1'},
              }

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
        }        

        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)

        print('Download bitstream')
        backend.run()

    def getInputKey(self):
        key = self.wire('user_pb', 4)
        self.addIn('user_pb', key)
        keyn = self.wire('user_pbn', 4)
        py4hw.Not(self, 'user_pbn', key, keyn)
        return keyn
        
    def getOutputLED(self):
        led = self.wire('user_led', 4)
        ledn = self.wire('user_ledn', 4)
        
        py4hw.Not(self, 'user_ledn', ledn, led)
        self.addOut('user_led', led)
        return ledn
    
    def getGPIO(self, connector:int):
        assert(False)
        from py4hw.external.altera.ip.cycloneiii import cycloneiii_io_buf
        name = f'GPIO{connector}_D'
        #wbidirs = self.wires(name + '_bidir', 32, 1)
        
        wbidirs = []
        
        for i in range(32):
            wbidirs.append( py4hw.FakeWire(f'{name}[{i}]'))
        
        self.addInOut(name, self.wire(name, 32))
        
        win = self.wire(name + '_in', 32) # out
        wout = self.wire(name + '_out', 32) # in
        woe = self.wire(name + '_oe', 32) # in
        
        wins = self.wires(name + '_in', 32, 1)
        wouts = self.wires(name + '_out', 32, 1)
        woes = self.wires(name + '_oe', 32, 1)
        
        for i in range(32):
            cycloneiii_io_buf(self, f'{name}_{i}', wins[i], wouts[i], woes[i], wbidirs[i])
        
        
        
        #py4hw.BitsLSBF(self, 'wins', win, wins)
        #py4hw.ConcatenateLSBF(self, 'wouts', wouts, wout)
        #py4hw.ConcatenateLSBF(self, 'woes', woes, woe)
        
        return [wins, wouts, woes]
    
    
    def getI2C(self):
        assert(False)
        fpga_i2c_sclk = self.addOut('FPGA_I2C_SCLK', self.wire('FPGA_I2C_SCLK'))
        fpga_i2c_sdat = self.addInOut('FPGA_I2C_SDAT', self.wire('FPGA_I2C_SDAT'))
        
        i2c = I2CInterface(self, 'fpga')
        
        py4hw.Buf(self, 'SCL', i2c.SCL, fpga_i2c_sclk)
        py4hw.BidirBuf(self, 'SDA', i2c.SDA_IN, i2c.SDA_OUT, i2c.SDA_OE, fpga_i2c_sdat)
        
        return i2c
    
    def getUART(self):
        #uart_txd = self.addOut('UART_TXD', self.wire('UART_TXD'))
        #uart_rxd = self.addIn('UART_RXD', self.wire('UART_RXD'))
        # uart_rts = self.addOut('UART_RTS', self.wire('UART_RTS'))
        # uart_cts = self.addIn('UART_CTS', self.wire('UART_CTS'))
        assert(False)
        uart = UARTInterface(self, 'uart')
        
        self.addInterfaceSource('', uart)
        
        return uart


    def getAudio(self):
        assert(False)
        audio = AudioInterface(self, 'audio')
        
        self.addInterfaceSource('', audio)
        
        return audio
    
    def getVGAController(self, vga_clk):
        assert(False)
        '''
        Connections to VGA connector
        

        Returns
        -------
        None.

        '''
        vga_ext = VGAExternalControllerInterface(self, 'DE0_EXT_VGA_CTRL', 4, useDAC=False)
        vga_int = VGAInternalControllerInterface(self, 'DE0_INT_VGA_CTRL')
        
        self.addInterfaceSource('VGA', vga_ext)
        
        py4hw.Range(self, 'R', vga_int.R, 7, 4, vga_ext.R)
        py4hw.Range(self, 'G', vga_int.G, 7, 4, vga_ext.G)
        py4hw.Range(self, 'B', vga_int.B, 7, 4, vga_ext.B)
        
        py4hw.Buf(self, 'HS', vga_int.HS, vga_ext.HS)
        py4hw.Buf(self, 'VS', vga_int.VS, vga_ext.VS)
        
        #py4hw.Buf(self, 'SYNC_N', vga_int.HS, vga_ext.SYNC_N)
        #py4hw.Buf(self, 'BLANK_N', vga_int.VS, vga_ext.BLANK_N)
        #py4hw.Buf(self, 'CLK', vga_clk, vga_ext.CLK)
        
        return vga_int
        
    def getFPGADeviceInfo(self):
        info = {'family':'"Cyclone 10 LP"', 'device':'10CL025YU256I7G',
                # 'package':'FBGA', 'pin_count':'896', 'speed_grade':'6'
                }
        return info
        
    def getPinAssignments(self):
        ##set_location_assignment PIN_C1  -to 'qspi_sdo     
        ##set_location_assignment PIN_D2  -to 'qspi_sce     
        ##set_location_assignment PIN_H1  -to 'qspi_dclk    
        ##set_location_assignment PIN_H2  -to 'qspi_data0   


        pins = {
            'c10_clk50m':'PIN_E1','c10_resetn':'PIN_J15',
            'enet_clk_125m':'PIN_T8','enet_rx_clk':'PIN_B8','enet_rx_dv':'PIN_A5',
            'enet_rx_d[0]':'PIN_A7','enet_rx_d[1]':'PIN_B7','enet_rx_d[2]':'PIN_A6','enet_rx_d[3]':'PIN_B6',
            'enet_tx_clk':'PIN_D3','enet_tx_en':'PIN_D6',
            'enet_tx_d[0]':'PIN_E6','enet_tx_d[1]':'PIN_A3','enet_tx_d[2]':'PIN_B3','enet_tx_d[3]':'PIN_A2',
            'enet_resetn':'PIN_C6','enet_mdc':'PIN_B4','enet_mdio':'PIN_A4','enet_int':'PIN_B5',
            'user_dip[0]':'PIN_M16','user_dip[1]':'PIN_A8','user_dip[2]':'PIN_A9','user_led[0]':'PIN_L14',
            'user_led[1]':'PIN_K15','user_led[2]':'PIN_J14','user_led[3]':'PIN_J13',
            'user_pb[0]':'PIN_E15','user_pb[1]':'PIN_F14','user_pb[2]':'PIN_C11','user_pb[3]':'PIN_D9',
            'arduino_io[0]':'PIN_B1','arduino_io[1]':'PIN_C2','arduino_io[2]':'PIN_F3','arduino_io[3]':'PIN_D1',
            'arduino_io[4]':'PIN_G2','arduino_io[5]':'PIN_G1','arduino_io[6]':'PIN_J2','arduino_io[7]':'PIN_J1',
            'arduino_io[8]':'PIN_K2','arduino_io[9]':'PIN_K5','arduino_io[10]':'PIN_L4','arduino_io[11]':'PIN_K1',
            'arduino_io[12]':'PIN_L2','arduino_io[13]':'PIN_L1','arduino_rstn':'PIN_L3','arduino_adc_sda':'PIN_C8',
            'arduino_adc_scl':'PIN_D8','arduino_scl':'PIN_N1','arduino_sda':'PIN_N2',
            'hbus_clk_50m':'PIN_M15','hbus_intn':'PIN_P11','hbus_rston':'PIN_T15',
            'hbus_dq[0]':'PIN_T12','hbus_dq[1]':'PIN_T13','hbus_dq[2]':'PIN_T11','hbus_dq[3]':'PIN_R10',
            'hbus_dq[4]':'PIN_T10','hbus_dq[5]':'PIN_R11','hbus_dq[6]':'PIN_R12','hbus_dq[7]':'PIN_R13',
            'hbus_cs2n':'PIN_P9','hbus_cs1n':'PIN_N12','hbus_rstn':'PIN_N9','hbus_rwds':'PIN_T14',
            'hbus_clk0p':'PIN_P14','hbus_clk0n':'PIN_R14',
            'c10_clk_adj':'PIN_E16',
            'c10_m10_io[0]':'PIN_E8','c10_m10_io[1]':'PIN_E7','c10_m10_io[2]':'PIN_F8','c10_m10_io[3]':'PIN_C3',
            'pmod_d[0]':'PIN_D16','pmod_d[1]':'PIN_F13','pmod_d[2]':'PIN_D15','pmod_d[3]':'PIN_F16',
            'pmod_d[4]':'PIN_C16','pmod_d[5]':'PIN_F15','pmod_d[6]':'PIN_C15','pmod_d[7]':'PIN_B16',
            'usb_data[0]':'PIN_A15','usb_data[1]':'PIN_B14','usb_data[2]':'PIN_A14','usb_data[3]':'PIN_B13',
            'usb_data[4]':'PIN_A13','usb_data[5]':'PIN_B12','usb_data[6]':'PIN_A12','usb_data[7]':'PIN_B10',
            'usb_addr[0]':'PIN_A10','usb_addr[1]':'PIN_A11',
            'usb_full':'PIN_D12','usb_empty':'PIN_D14',
            'usb_scl':'PIN_C14','usb_sda':'PIN_E9',
            'usb_oe_n':'PIN_F9','usb_rd_n':'PIN_D11','usb_wr_n':'PIN_B11',
            'usb_reset_n':'PIN_C9','c10_usb_clk':'PIN_B9',
            'gpio[0]':'PIN_L13','gpio[1]':'PIN_L16','gpio[2]':'PIN_L15','gpio[3]':'PIN_K16',
            'gpio[4]':'PIN_P16','gpio[5]':'PIN_R16','gpio[6]':'PIN_N16','gpio[7]':'PIN_N15',
            'gpio[8]':'PIN_N14','gpio[9]':'PIN_P15','gpio[10]':'PIN_N8','gpio[11]':'PIN_P8',
            'gpio[12]':'PIN_M8','gpio[13]':'PIN_L8','gpio[14]':'PIN_R7','gpio[15]':'PIN_T7',
            'gpio[16]':'PIN_L7','gpio[17]':'PIN_M7','gpio[18]':'PIN_R6','gpio[19]':'PIN_T6',
            'gpio[20]':'PIN_T2','gpio[21]':'PIN_M6','gpio[22]':'PIN_R5','gpio[23]':'PIN_T5',
            'gpio[24]':'PIN_N5','gpio[25]':'PIN_N6','gpio[26]':'PIN_R4','gpio[27]':'PIN_T4',
            'gpio[28]':'PIN_N3','gpio[29]':'PIN_P3','gpio[30]':'PIN_R3','gpio[31]':'PIN_T3',
            'gpio[32]':'PIN_P6','gpio[33]':'PIN_P2','gpio[34]':'PIN_P1','gpio[35]':'PIN_R1',                
            }
        return pins
        
    def getQsf(self):
        info = self.getFPGADeviceInfo()
        qsf = '# build by py4hw\n'
        qsf += 'set_global_assignment -name FAMILY {}\n'.format(info['family'])
        qsf += 'set_global_assignment -name DEVICE {}\n'.format(info['device'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_PACKAGE {}\n'.format(info['package'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_PIN_COUNT {}\n'.format(info['pin_count'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_SPEED_GRADE {}\n'.format(info['speed_grade'])

        pins = self.getPinAssignments()
        
        for key in pins.keys():
            qsf += 'set_location_assignment {} -to {}\n'.format(pins[key], key)

        # by default all are 3.3 LVTTL
        lvpins = ['hbus_intn','hbus_rston','hbus_dq[0]','hbus_dq[1]',
                  'hbus_dq[2]','hbus_dq[3]','hbus_dq[4]','hbus_dq[5]',
                  'hbus_dq[6]','hbus_dq[7]','hbus_cs2n','hbus_cs1n',
                  'hbus_rstn','hbus_rwds','hbus_clk0p','hbus_clk0n']

        for key in pins.keys():
            if key in lvpins:
                qsf += 'set_instance_assignment -name IO_STANDARD "1.8 V" -to {}\n'.format(key)
            else:
                qsf += 'set_instance_assignment -name IO_STANDARD "3.3-V LVTTL" -to {}\n'.format(key)

        qsf += 'set_global_assignment -name ENABLE_OCT_DONE OFF'
        qsf += 'set_global_assignment -name ENABLE_CONFIGURATION_PINS OFF'
        qsf += 'set_global_assignment -name ENABLE_BOOT_SEL_PIN OFF'
        qsf += 'set_global_assignment -name STRATIXV_CONFIGURATION_SCHEME "PASSIVE SERIAL"'
        qsf += 'set_global_assignment -name USE_CONFIGURATION_DEVICE OFF'
        qsf += 'set_global_assignment -name CRC_ERROR_OPEN_DRAIN OFF'
        qsf += 'set_global_assignment -name CYCLONEII_RESERVE_NCEO_AFTER_CONFIGURATION "USE AS REGULAR IO"\n'
        
        return qsf



# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 17:39:10 2024

@author: 2016570
"""

import py4hw
import math

class xilinx_xpm_fifo_sync(py4hw.Logic):
    
    def __init__(self, parent, name, reset, wr_en, rd_en, din, dout, empty, full, prog_empty, prog_full,
                 FIFO_DEPTH, PROG_FULL_THRESH, PROG_EMPTY_THRESH):
        
        super().__init__(parent, name)
        
        self.addIn('reset', reset)
        self.addIn('wr_en', wr_en)
        self.addIn('rd_en', rd_en)
        self.addIn('din', din)
        
        self.addOut('dout', dout)
        self.addOut('empty', empty)
        self.addOut('full', full)
        
        self.addOut('prog_empty', prog_empty)
        self.addOut('prog_full', prog_full)
        
        self.FIFO_DEPTH = FIFO_DEPTH
        self.PROG_FULL_THRESH = PROG_FULL_THRESH
        self.PROG_EMPTY_THRESH = PROG_EMPTY_THRESH
        self.dw = din.getWidth()
        
    def clock(self):
        # No behavioural model yet
        pass
    
    def verilogBody(self):
        str = ''

        from py4hw.base import getObjectClockDriver            
        clk_name = getObjectClockDriver(self).name

        str += '// Xilinx Parameterized Macro, Version 2016.4\n'
        str += 'xpm_fifo_sync # (\n'
        str += '.FIFO_MEMORY_TYPE("auto"),           //string; "auto", "block", "distributed", or "ultra";\n'
        str += '.ECC_MODE("no_ecc"),         //string; "no_ecc" or "en_ecc";\n'
        str += f'.FIFO_WRITE_DEPTH({self.FIFO_DEPTH}),   //positive integer\n'
        str += '.WRITE_DATA_WIDTH({}),        //positive integer\n'.format(self.dw)
        str += '.WR_DATA_COUNT_WIDTH({}),       //positive integer, Not used\n'.format(int(math.ceil(math.log2(self.FIFO_DEPTH)))+1) # @todo this +1 does not make sense
        str += '.PROG_FULL_THRESH({}), //positive integer\n'.format(self.PROG_FULL_THRESH)
        str += '.FULL_RESET_VALUE(1), //positive integer; 0 or 1\n'        
        str += '.READ_MODE("fwft"), //string; "std" or "fwft";\n'
        str += '.FIFO_READ_LATENCY(1), //positive integer;\n'
        str += '.READ_DATA_WIDTH({}), //positive integer\n'.format(self.dw)
        str += '.RD_DATA_COUNT_WIDTH({}), //positive integer, not used\n'.format(int(math.ceil(math.log2(self.FIFO_DEPTH)))+1)# @todo this +1 does not make sense
        str += '.PROG_EMPTY_THRESH({}),               //positive integer\n'.format(self.PROG_EMPTY_THRESH)
        str += '.DOUT_RESET_VALUE          ("0"),              //string,\n'
        str += '.WAKEUP_TIME               (0)                 //positive integer; 0 or 2;\n'
        
        str += ') inst_rd_xpm_fifo_sync (\n'
        str += ".sleep(1'b0), .rst( reset) ,\n"
        str += f'.wr_clk        ( {clk_name}) ,\n'
        str += '.wr_en(wr_en), .din(din), .full(full), .prog_full(prog_full),\n'
        
        str += '''        
  .wr_data_count (                  ) ,
  .overflow      (                  ) ,
  .wr_rst_busy   (                  ) ,
  .rd_en         ( rd_en   ) ,
  .dout          ( dout    ) ,
  .empty         ( empty ) ,
  .prog_empty    ( prog_empty ) ,
  .rd_data_count (                  ) ,
  .underflow     (                  ) ,
  .rd_rst_busy   (                  ) ,
  .injectsbiterr ( 1'b0             ) ,
  .injectdbiterr ( 1'b0             ) ,
  .sbiterr       (                  ) ,
  .dbiterr       (                  ) 

);
        '''
        
        return str


# -*- coding: utf-8 -*-
"""
Created on Sun Apr  2 14:23:01 2023

@author: dcr
"""

from .. import *


# From                              |  To                               | Class                                            | Instance / Class
# Structural py4hw                  | Verilog                           | VerilogGenerator                                 | Instance
# Behavioural combinational py4hw   | Verilog                           | VerilogGenerator -> python2verilog_transpiler    | Instance
# Behavioural sequential py4hw      | Verilog                           | VerilogGenerator -> python2verilog_transpiler    | Instance
# Algorithmic Timed py4hw           | Verilog                           | ?
# Behavioural combinational py4hw   | Structuural py4hw                 | ?
# Behavioural combinational py4hw   | Structuural py4hw                 | ? 
# Algorithmic Timed py4hw           | Behavioural combinational py4hw   | Py4hwGenerator
 


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  8 15:12:40 2024

@author: 2016570
"""

import py4hw

class alt_io_buf (py4hw.Logic):
    
    def __init__(self, parent, name, pin, pout, poe, bidir):
        
        super().__init__(parent, name)
        
        assert(poe.getWidth() == 1)
        
        self.pin = self.addOut('pin', pin)
        self.pout = self.addIn('pout', pout)
        self.poe = self.addIn('poe', poe)
        
        self.bidir = self.addInOut('bidit', bidir)
        
    def structureName(self):
        return 'alt_io_buf_{}'.format(self.bidir.getWidth())
    
    def propagate(self):
        pass
    
    def verilogBody(self):
        ret = ''
        
        ret += f'ALTIOBUF i0 (\n'
        ret += '.i(pout),      // Data output from FPGA\n'
        ret += '.oe(poe),      // Output enable signal\n'
        ret += '.io(bidir),    // Bidirectional pin\n'
        ret += '.o(pin));      // Data input to FPGA\n'
        return ret



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  4 19:34:59 2023

@author: dcr
"""
import py4hw

class intel_lpm_counter(py4hw.Logic):
    def __init__(self, parent, name, reset, q):
        super().__init__(parent, name)
        
        if (reset):
            self.reset = self.addIn('reset', reset)
        else:
            self.reset = None
            
        self.q = self.addOut('q', q)
        
    def clock(self):
        if not(self.reset is None):
            if (self.reset.get() == 1):
                self.q.prepare(0)
                return

        self.q.prepare(self.q.get() + 1)

    def verilogBody(self):
        
        clkname = py4hw.getObjectClockDriver(self).name
        
        str = ' lpm_counter i0 ('
        str += '  .clock('+clkname+'),\n'
        str += '  .q(q),\n'
        
        if not(self.reset is None):
            str += '  .sclr(reset),\n'

        str += "  .updown(1'b1));\n"
        str += ' defparam\n'
        str += '   i0.lpm_direction = "UP",\n'
        str += '   i0.lpm_port_updown = "PORT_UNUSED",\n'
        str += '   i0.lpm_type = "LPM_COUNTER",\n'
        str += '   i0.lpm_width = {};\n'.format(self.q.getWidth())

        return str

#  lpm_counter	LPM_COUNTER_component (
# .clock (CLOCK_50),
# .q (sub_wire0),
# .aclr (1'b0),
# .aload (1'b0),
# .aset (1'b0),
# .cin (1'b1),
# .clk_en (1'b1),
# .cnt_en (1'b1),
# .cout (),
# .data ({64{1'b0}}),
# .eq (),
# .sclr (1'b0),
# .sload (1'b0),
# .sset (1'b0),
# .updown (1'b1));
# param
# M_COUNTER_component.lpm_direction = "UP",
# M_COUNTER_component.lpm_port_updown = "PORT_UNUSED",
# M_COUNTER_component.lpm_type = "LPM_COUNTER",
# M_COUNTER_component.lpm_width = 64;



# -*- coding: utf-8 -*-
"""
Created on Sat Sep 30 09:01:29 2023

@author: dcr
"""



import py4hw
from edalize import edatool
import os

class VGAExternalControllerInterface(py4hw.Interface):
    def __init__(self, parent, name:str, w=8, useDAC=True):
        super().__init__(parent, name)
        
        self.R = self.addSourceToSink('R', w)
        self.G = self.addSourceToSink('G', w)
        self.B = self.addSourceToSink('B', w)
        
        if (useDAC):
            self.CLK = self.addSourceToSink('CLK', 1)
            self.SYNC_N = self.addSourceToSink('SYNC_N', 1)
            self.BLANK_N = self.addSourceToSink('BLANK_N', 1)

        self.VS = self.addSourceToSink('VS', 1)
        self.HS = self.addSourceToSink('HS', 1)
        
class VGAInternalControllerInterface(py4hw.Interface):
    def __init__(self, parent, name:str):
        super().__init__(parent, name)
        
        self.R = self.addSourceToSink('R', 8)
        self.G = self.addSourceToSink('G', 8)
        self.B = self.addSourceToSink('B', 8)
        
        self.VS = self.addSourceToSink('VS', 1)
        self.HS = self.addSourceToSink('HS', 1)
        
class I2CInterface(py4hw.Interface):
    def __init__(self, parent, name:str):
        super().__init__(parent, name)
        
        self.SCL = self.addSourceToSink('SCL', 1)
        self.SDA_OUT = self.addSourceToSink('SDA_OUT', 1)
        self.SDA_OE = self.addSourceToSink('SDA_OE', 1)
        self.SDA_IN = self.addSinkToSource('SDA_IN', 1)

class UARTInterface(py4hw.Interface):
    def __init__(self, parent, name:str):
        super().__init__(parent, name)
        
        self.tx = self.addSourceToSink('UART_TXD', 1)
        self.rx = self.addSinkToSource('UART_RXD', 1)
        self.rts = self.addSourceToSink('UART_RTS', 1)
        self.cts = self.addSinkToSource('UART_CTS', 1)


class AudioInterface(py4hw.Interface):
    def __init__(self, parent, name:str):
        super().__init__(parent, name)
        
        self.AUD_ADCLRCK = self.addSourceToSink('AUD_ADCLRCK', 1)
        self.AUD_ADCDAT =  self.addSinkToSource('AUD_ADCDAT', 1)
        self.AUD_DACLRCK = self.addSourceToSink('AUD_DACLRCK', 1)
        self.AUD_DACDAT = self.addSourceToSink('AUD_DACDAT', 1)
        self.AUD_XCK = self.addSourceToSink('AUD_XCK', 1)
        self.AUD_BCLK = self.addSourceToSink('AUD_BCLK', 1)

    
class DE0(py4hw.HWSystem):
    
    def __init__(self):
        
        super().__init__(name='DE0')
        
        clk50 = self.wire('CLOCK_50')
        #clk50 = self.addIn('CLOCK_50', clk50)
        clockDriver = py4hw.ClockDriver('CLOCK_50', 50E6, 0, wire=clk50)
        
        print('Initial', self.getFullPath(),  self._wires)
        
        self.clockDriver = clockDriver
    
    
    
    def build(self,  projectDir, createdStructures=[]):
        if not(os.path.exists(projectDir)):
            print('Creating the directory', projectDir)
            os.makedirs(projectDir)
            
        rtl = py4hw.VerilogGenerator(self)
        
        rtl_code = rtl.getVerilogForHierarchy(noInstanceNumberInTopEntity=True, createdStructures=createdStructures)
        qsf_code = self.getQsf()
        top_name = self.name
        
        verilog_file = os.path.join(projectDir, top_name+'.v')
        qsf_file = os.path.join(projectDir, top_name+'_pins.qsf')
        sdc_file = os.path.join(projectDir, 'CLOCK_50.sdc')

        # Create Verilog file
        with open(verilog_file, 'w') as file:
            file.write(rtl_code)

        # Create Verilog file
        with open(qsf_file, 'w') as file:
            file.write(qsf_code)
            
        with open(sdc_file, 'w') as file:
            file.write('create_clock  -period 20 [get_ports CLOCK_50]\n')
            
        tool = 'quartus'
        
        files = [  {'name' : verilog_file, 'file_type' : 'verilogSource'},
                 {'name' : sdc_file, 'file_type' : 'SDC'},
                 {'name' : qsf_file, 'file_type' : 'tclSource'}
                 ]

        info = self.getFPGADeviceInfo()

        parameters = {'clk_freq_hz' : {'datatype' : 'int', 'default' : 50000000, 'paramtype' : 'vlogparam'},
              'vcd' : {'datatype' : 'bool', 'paramtype' : 'plusarg'},
              'family': {'datatype': 'string', 'paramtype': 'generic', 'default':'Cyclone III'},
              'device': {'datatype': 'string', 'paramtype': 'generic', 'default':info['device']},
              }

        tool_options = {'family':info['family'],
                        'flags': ['--64bit'],
                        'quartus': {'flags': ['--64bit']}
                        }

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
          'tool_options' : tool_options
        }        
        
        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)
        
        print('Configure Edalize')
        backend.configure()
        
        print('Build Edalize')
        backend.build()
    
    def download(self, projectDir):
        tool='quartus'
        top_name = self.name
        
        files=[]
        parameters={
              'cable': {'datatype': 'string', 'paramtype': 'generic', 'default':'USB-Blaster'},
              'board_device_index': {'datatype': 'string', 'paramtype': 'generic', 'default':'1'},
              }

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
        }        

        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)

        print('Download bitstream')
        backend.run()

    def getInputKey(self):
        key = self.wire('BUTTON', 3)
        self.addIn('BUTTON', key)
        keyn = self.wire('BUTTONn', 3)
        py4hw.Not(self, 'BUTTONn', key, keyn)
        return keyn
        
    def getOutputHex(self, i):
        assert(i >= 0 and i <= 5)
        name = 'HEX{}_D'.format(i)
        namen = 'HEXn{}'.format(i)
        p = self.wire(name, 7)
        pn = self.wire(namen, 7)
        
        py4hw.Not(self, namen, pn, p)
        self.addOut(name, p)
        return pn
    
    def getGPIO(self, connector:int):
        from py4hw.external.altera.ip.cycloneiii import cycloneiii_io_buf
        name = f'GPIO{connector}_D'
        #wbidirs = self.wires(name + '_bidir', 32, 1)
        
        wbidirs = []
        
        for i in range(32):
            wbidirs.append( py4hw.FakeWire(f'{name}[{i}]'))
        
        self.addInOut(name, self.wire(name, 32))
        
        win = self.wire(name + '_in', 32) # out
        wout = self.wire(name + '_out', 32) # in
        woe = self.wire(name + '_oe', 32) # in
        
        wins = self.wires(name + '_in', 32, 1)
        wouts = self.wires(name + '_out', 32, 1)
        woes = self.wires(name + '_oe', 32, 1)
        
        for i in range(32):
            cycloneiii_io_buf(self, f'{name}_{i}', wins[i], wouts[i], woes[i], wbidirs[i])
        
        
        
        #py4hw.BitsLSBF(self, 'wins', win, wins)
        #py4hw.ConcatenateLSBF(self, 'wouts', wouts, wout)
        #py4hw.ConcatenateLSBF(self, 'woes', woes, woe)
        
        return [wins, wouts, woes]
    
    
    def getI2C(self):
        fpga_i2c_sclk = self.addOut('FPGA_I2C_SCLK', self.wire('FPGA_I2C_SCLK'))
        fpga_i2c_sdat = self.addInOut('FPGA_I2C_SDAT', self.wire('FPGA_I2C_SDAT'))
        
        i2c = I2CInterface(self, 'fpga')
        
        py4hw.Buf(self, 'SCL', i2c.SCL, fpga_i2c_sclk)
        py4hw.BidirBuf(self, 'SDA', i2c.SDA_IN, i2c.SDA_OUT, i2c.SDA_OE, fpga_i2c_sdat)
        
        return i2c
    
    def getUART(self):
        #uart_txd = self.addOut('UART_TXD', self.wire('UART_TXD'))
        #uart_rxd = self.addIn('UART_RXD', self.wire('UART_RXD'))
        # uart_rts = self.addOut('UART_RTS', self.wire('UART_RTS'))
        # uart_cts = self.addIn('UART_CTS', self.wire('UART_CTS'))
        
        uart = UARTInterface(self, 'uart')
        
        self.addInterfaceSource('', uart)
        
        return uart


    def getAudio(self):
        
        audio = AudioInterface(self, 'audio')
        
        self.addInterfaceSource('', audio)
        
        return audio
    
    def getVGAController(self, vga_clk):
        '''
        Connections to VGA connector
        

        Returns
        -------
        None.

        '''
        vga_ext = VGAExternalControllerInterface(self, 'DE0_EXT_VGA_CTRL', 4, useDAC=False)
        vga_int = VGAInternalControllerInterface(self, 'DE0_INT_VGA_CTRL')
        
        self.addInterfaceSource('VGA', vga_ext)
        
        py4hw.Range(self, 'R', vga_int.R, 7, 4, vga_ext.R)
        py4hw.Range(self, 'G', vga_int.G, 7, 4, vga_ext.G)
        py4hw.Range(self, 'B', vga_int.B, 7, 4, vga_ext.B)
        
        py4hw.Buf(self, 'HS', vga_int.HS, vga_ext.HS)
        py4hw.Buf(self, 'VS', vga_int.VS, vga_ext.VS)
        
        #py4hw.Buf(self, 'SYNC_N', vga_int.HS, vga_ext.SYNC_N)
        #py4hw.Buf(self, 'BLANK_N', vga_int.VS, vga_ext.BLANK_N)
        #py4hw.Buf(self, 'CLK', vga_clk, vga_ext.CLK)
        
        return vga_int
        
    def getFPGADeviceInfo(self):
        info = {'family':'"Cyclone III"', 'device':'EP3C16F484C6',
                # 'package':'FBGA', 'pin_count':'896', 'speed_grade':'6'
                }
        return info
        
    def getPinAssignments(self):
        pins = {'LEDG[9]':'PIN_B1', 'LEDG[8]':'PIN_B2', 'LEDG[7]':'PIN_C2', 'LEDG[6]':'PIN_C1', 'LEDG[5]':'PIN_E1', 
                'LEDG[4]':'PIN_F2', 'LEDG[3]':'PIN_H1', 'LEDG[2]':'PIN_J3', 'LEDG[1]':'PIN_J2', 'LEDG[0]':'PIN_J1', 
                'SW[9]':'PIN_D2', 'SW[8]':'PIN_E4', 'SW[7]':'PIN_E3', 'SW[6]':'PIN_H7', 'SW[5]':'PIN_J7', 
                'SW[4]':'PIN_G5', 'SW[3]':'PIN_G4', 'SW[2]':'PIN_H6', 'SW[1]':'PIN_H5', 'SW[0]':'PIN_J6', 
                'BUTTON[2]':'PIN_F1', 'BUTTON[1]':'PIN_G3', 'BUTTON[0]':'PIN_H2',
                'FL_ADDR[21]':'PIN_R2', 'FL_ADDR[20]':'PIN_P3', 'FL_ADDR[19]':'PIN_P1', 'FL_ADDR[18]':'PIN_M6', 'FL_ADDR[17]':'PIN_M5',
                'FL_ADDR[16]':'PIN_AA2', 'FL_ADDR[15]':'PIN_L6', 'FL_ADDR[14]':'PIN_L7', 'FL_ADDR[13]':'PIN_M1', 'FL_ADDR[12]':'PIN_M2',
                'FL_ADDR[11]':'PIN_M3', 'FL_ADDR[10]':'PIN_N1', 'FL_ADDR[9]':'PIN_N2', 'FL_ADDR[8]':'PIN_P2', 'FL_ADDR[7]':'PIN_M4',
                'FL_ADDR[6]':'PIN_M8', 'FL_ADDR[5]':'PIN_N6', 'FL_ADDR[4]':'PIN_N5', 'FL_ADDR[3]':'PIN_N7', 'FL_ADDR[2]':'PIN_P6',
                'FL_ADDR[1]':'PIN_P5', 'FL_ADDR[0]':'PIN_P7', 'FL_BYTE_N':'PIN_AA1', 'FL_CE_N':'PIN_N8', 
                'FL_DQ[0]':'PIN_R7', 'FL_DQ[1]':'PIN_P8', 'FL_DQ[2]':'PIN_R8', 'FL_DQ[3]':'PIN_U1', 'FL_DQ[4]':'PIN_V2',
                'FL_DQ[5]':'PIN_V3', 'FL_DQ[6]':'PIN_W1', 'FL_DQ[7]':'PIN_Y1', 'FL_DQ[8]':'PIN_T5', 'FL_DQ[9]':'PIN_T7',
                'FL_DQ[10]':'PIN_T4', 'FL_DQ[11]':'PIN_U2', 'FL_DQ[12]':'PIN_V1', 'FL_DQ[13]':'PIN_V4', 'FL_DQ[14]':'PIN_W2', 
                'FL_OE_N':'PIN_R6', 'FL_RST_N':'PIN_R1', 'FL_RY':'PIN_M7', 'FL_WE_N':'PIN_P4', 'FL_WP_N':'PIN_T3', 'FL_DQ15_AM1':'PIN_Y2',
                'GPIO0_D[31]':'PIN_U7', 'GPIO0_D[30]':'PIN_V5', 'GPIO0_D[29]':'PIN_W6', 'GPIO0_D[28]':'PIN_W7',
                'GPIO0_D[27]':'PIN_V8', 'GPIO0_D[26]':'PIN_T8', 'GPIO0_D[25]':'PIN_W10', 'GPIO0_D[24]':'PIN_Y10',
                'GPIO0_D[23]':'PIN_V11', 'GPIO0_D[22]':'PIN_R10', 'GPIO0_D[21]':'PIN_V12', 'GPIO0_D[20]':'PIN_U13',
                'GPIO0_D[19]':'PIN_W13', 'GPIO0_D[18]':'PIN_Y13', 'GPIO0_D[17]':'PIN_U14', 'GPIO0_D[16]':'PIN_V14',
                'GPIO0_D[15]':'PIN_AA4', 'GPIO0_D[14]':'PIN_AB4', 'GPIO0_D[13]':'PIN_AA5', 'GPIO0_D[12]':'PIN_AB5',
                'GPIO0_D[11]':'PIN_AA8', 'GPIO0_D[10]':'PIN_AB8', 'GPIO0_D[9]':'PIN_AA10', 'GPIO0_D[8]':'PIN_AB10',
                'GPIO0_D[7]':'PIN_AA13', 'GPIO0_D[6]':'PIN_AB13', 'GPIO0_D[5]':'PIN_AB14', 'GPIO0_D[4]':'PIN_AA14',
                'GPIO0_D[3]':'PIN_AB15', 'GPIO0_D[2]':'PIN_AA15', 'GPIO0_D[1]':'PIN_AA16', 'GPIO0_D[0]':'PIN_AB16',
                'GPIO0_CLKIN[0]':'PIN_AB12', 'GPIO0_CLKIN[1]':'PIN_AA12', 'GPIO0_CLKOUT[0]':'PIN_AB3', 'GPIO0_CLKOUT[1]':'PIN_AA3',
                'GPIO1_CLKIN[1]':'PIN_AA11', 'GPIO1_CLKIN[0]':'PIN_AB11', 'GPIO1_CLKOUT[1]':'PIN_T16', 'GPIO1_CLKOUT[0]':'PIN_R16',
                'GPIO1_D[31]':'PIN_V7', 'GPIO1_D[30]':'PIN_V6', 'GPIO1_D[29]':'PIN_U8', 'GPIO1_D[28]':'PIN_Y7',
                'GPIO1_D[27]':'PIN_T9', 'GPIO1_D[26]':'PIN_U9', 'GPIO1_D[25]':'PIN_T10', 'GPIO1_D[24]':'PIN_U10',
                'GPIO1_D[23]':'PIN_R12', 'GPIO1_D[22]':'PIN_R11', 'GPIO1_D[21]':'PIN_T12', 'GPIO1_D[20]':'PIN_U12',
                'GPIO1_D[19]':'PIN_R14', 'GPIO1_D[18]':'PIN_T14', 'GPIO1_D[17]':'PIN_AB7', 'GPIO1_D[16]':'PIN_AA7',
                'GPIO1_D[15]':'PIN_AA9', 'GPIO1_D[14]':'PIN_AB9', 'GPIO1_D[13]':'PIN_V15', 'GPIO1_D[12]':'PIN_W15', 
                'GPIO1_D[11]':'PIN_T15', 'GPIO1_D[10]':'PIN_U15', 'GPIO1_D[9]':'PIN_W17', 'GPIO1_D[8]':'PIN_Y17',
                'GPIO1_D[7]':'PIN_AB17', 'GPIO1_D[6]':'PIN_AA17', 'GPIO1_D[5]':'PIN_AA18', 'GPIO1_D[4]':'PIN_AB18',
                'GPIO1_D[3]':'PIN_AB19', 'GPIO1_D[2]':'PIN_AA19', 'GPIO1_D[1]':'PIN_AB20', 'GPIO1_D[0]':'PIN_AA20',
                'PS2_KBCLK':'PIN_P22', 'PS2_KBDAT':'PIN_P21', 'PS2_MSCLK':'PIN_R21', 'PS2_MSDAT':'PIN_R22',
                'UART_RXD':'PIN_U22', 'UART_TXD':'PIN_U21', 'UART_RTS':'PIN_V22', 'UART_CTS':'PIN_V21',
                'SD_CLK':'PIN_Y21', 'SD_CMD':'PIN_Y22', 'SD_DAT0':'PIN_AA22', 'SD_DAT3':'PIN_W21', 'SD_WP_N':'PIN_W20',
                'LCD_DATA[7]':'PIN_C20', 'LCD_DATA[6]':'PIN_D20', 'LCD_DATA[5]':'PIN_B21', 'LCD_DATA[4]':'PIN_B22',
                'LCD_DATA[3]':'PIN_C21', 'LCD_DATA[2]':'PIN_C22', 'LCD_DATA[1]':'PIN_D21', 'LCD_DATA[0]':'PIN_D22',
                'LCD_RW':'PIN_E22', 'LCD_RS':'PIN_F22', 'LCD_EN':'PIN_E21', 'LCD_BLON':'PIN_F21',
                'VGA_G[3]':'PIN_J21', 'VGA_G[2]':'PIN_K17', 'VGA_G[1]':'PIN_J17', 'VGA_G[0]':'PIN_H22', 
                'VGA_HS':'PIN_L21', 'VGA_VS':'PIN_L22',
                'VGA_R[3]':'PIN_H21', 'VGA_R[2]':'PIN_H20', 'VGA_R[1]':'PIN_H17', 'VGA_R[0]':'PIN_H19',
                'VGA_B[3]':'PIN_K18', 'VGA_B[2]':'PIN_J22', 'VGA_B[1]':'PIN_K21', 'VGA_B[0]':'PIN_K22',
                'CLOCK_50':'PIN_G21',
                'HEX0_D[0]':'PIN_E11', 'HEX0_D[1]':'PIN_F11', 'HEX0_D[2]':'PIN_H12', 'HEX0_D[3]':'PIN_H13', 
                'HEX0_D[4]':'PIN_G12', 'HEX0_D[5]':'PIN_F12', 'HEX0_D[6]':'PIN_F13', 'HEX0_DP':'PIN_D13',
                'HEX1_D[6]':'PIN_A15', 'HEX1_D[5]':'PIN_E14', 'HEX1_D[4]':'PIN_B14', 'HEX1_D[3]':'PIN_A14',
                'HEX1_D[2]':'PIN_C13', 'HEX1_D[1]':'PIN_B13', 'HEX1_D[0]':'PIN_A13', 'HEX1_DP':'PIN_B15',
                'HEX2_D[6]':'PIN_F14', 'HEX2_D[5]':'PIN_B17', 'HEX2_D[4]':'PIN_A17', 'HEX2_D[3]':'PIN_E15',
                'HEX2_D[2]':'PIN_B16', 'HEX2_D[1]':'PIN_A16', 'HEX2_D[0]':'PIN_D15', 'HEX2_DP':'PIN_A18',
                'HEX3_D[6]':'PIN_G15', 'HEX3_D[5]':'PIN_D19', 'HEX3_D[4]':'PIN_C19', 'HEX3_D[3]':'PIN_B19',
                'HEX3_D[2]':'PIN_A19', 'HEX3_D[1]':'PIN_F15', 'HEX3_D[0]':'PIN_B18', 'HEX3_DP':'PIN_G16',   
                'DRAM_CAS_N':'PIN_G8', 'DRAM_CS_N':'PIN_G7', 'DRAM_CLK':'PIN_E5', 'DRAM_CKE':'PIN_E6',
                'DRAM_BA_0':'PIN_B5', 'DRAM_BA_1':'PIN_A4', 'DRAM_DQ[15]':'PIN_F10', 'DRAM_DQ[14]':'PIN_E10',
                'DRAM_DQ[13]':'PIN_A10', 'DRAM_DQ[12]':'PIN_B10', 'DRAM_DQ[11]':'PIN_C10', 'DRAM_DQ[10]':'PIN_A9',
                'DRAM_DQ[9]':'PIN_B9', 'DRAM_DQ[8]':'PIN_A8', 'DRAM_DQ[7]':'PIN_F8', 'DRAM_DQ[6]':'PIN_H9',
                'DRAM_DQ[5]':'PIN_G9', 'DRAM_DQ[4]':'PIN_F9', 'DRAM_DQ[3]':'PIN_E9', 'DRAM_DQ[2]':'PIN_H10',
                'DRAM_DQ[1]':'PIN_G10', 'DRAM_DQ[0]':'PIN_D10', 'DRAM_LDQM':'PIN_E7', 'DRAM_UDQM':'PIN_B8',
                'DRAM_RAS_N':'PIN_F7', 'DRAM_WE_N':'PIN_D6', 'CLOCK_50_2':'PIN_B12', 'DRAM_ADDR[12]':'PIN_C8',
                'DRAM_ADDR[11]':'PIN_A7', 'DRAM_ADDR[10]':'PIN_B4', 'DRAM_ADDR[9]':'PIN_B7', 'DRAM_ADDR[8]':'PIN_C7',
                'DRAM_ADDR[7]':'PIN_A6', 'DRAM_ADDR[6]':'PIN_B6', 'DRAM_ADDR[5]':'PIN_C6', 'DRAM_ADDR[4]':'PIN_A5',
                'DRAM_ADDR[3]':'PIN_C3', 'DRAM_ADDR[2]':'PIN_B3', 'DRAM_ADDR[1]':'PIN_A3', 'DRAM_ADDR[0]':'PIN_C4',
                }
        return pins
        
    def getQsf(self):
        info = self.getFPGADeviceInfo()
        qsf = '# build by py4hw\n'
        qsf += 'set_global_assignment -name FAMILY {}\n'.format(info['family'])
        qsf += 'set_global_assignment -name DEVICE {}\n'.format(info['device'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_PACKAGE {}\n'.format(info['package'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_PIN_COUNT {}\n'.format(info['pin_count'])
        # qsf += 'set_global_assignment -name DEVICE_FILTER_SPEED_GRADE {}\n'.format(info['speed_grade'])

        pins = self.getPinAssignments()
        
        for key in pins.keys():
            qsf += 'set_location_assignment {} -to {}\n'.format(pins[key], key)

        # by default all are 3.3 LVTTL
        for key in pins.keys():
            qsf += 'set_instance_assignment -name IO_STANDARD "3.3-V LVTTL" -to {}\n'.format(key)
            
        qsf += 'set_global_assignment -name CYCLONEII_RESERVE_NCEO_AFTER_CONFIGURATION "USE AS REGULAR IO"\n'

        
        return qsf
        

    
class DE1SoC(py4hw.HWSystem):
    
    def __init__(self):
        
        super().__init__(name='DE1SoC')
        
        self.clk_freq = 50E6
        clk50 = self.wire('CLOCK_50')
        #clk50 = self.addIn('CLOCK_50', clk50)
        clockDriver = py4hw.ClockDriver('CLOCK_50', self.clk_freq , 0, wire=clk50)
        
        print('Initial', self.getFullPath(),  self._wires)
        
        self.clockDriver = clockDriver
        
    
    def setTargetClkFreq(self, freq):
        self.clk_freq = freq
    
    def build(self,  projectDir):
        if not(os.path.exists(projectDir)):
            print('Creating the directory', projectDir)
            os.makedirs(projectDir)
            
        rtl = py4hw.VerilogGenerator(self)
        
        rtl_code = rtl.getVerilogForHierarchy(noInstanceNumberInTopEntity=True)
        qsf_code = self.getQsf()
        top_name = self.name
        
        verilog_file = os.path.join(projectDir, top_name+'.v')
        qsf_file = os.path.join(projectDir, top_name+'_pins.qsf')
        sdc_file = os.path.join(projectDir, 'CLOCK_50.sdc')

        # Create Verilog file
        with open(verilog_file, 'w') as file:
            file.write(rtl_code)

        # Create Verilog file
        with open(qsf_file, 'w') as file:
            file.write(qsf_code)
            
        with open(sdc_file, 'w') as file:
            period_in_ns = 1E9 / self.clk_freq
            file.write('create_clock  -period {} [get_ports CLOCK_50]\n'.format(period_in_ns))
            
        tool = 'quartus'
        
        files = [  {'name' : verilog_file, 'file_type' : 'verilogSource'},
                 {'name' : sdc_file, 'file_type' : 'SDC'},
                 {'name' : qsf_file, 'file_type' : 'tclSource'}
                 ]

        parameters = {'clk_freq_hz' : {'datatype' : 'int', 'default' : int(self.clk_freq), 'paramtype' : 'vlogparam'},
              'vcd' : {'datatype' : 'bool', 'paramtype' : 'plusarg'},
              'family': {'datatype': 'string', 'paramtype': 'generic', 'default':'Cyclone V'},
              'device': {'datatype': 'string', 'paramtype': 'generic', 'default':'5CSEMA5F31C6'},
              }

        tool_options = {'family':'Cyclone V'}

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
          'tool_options' : tool_options
        }        
        
        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)
        
        print('Configure Edalize')
        backend.configure()
        
        print('Build Edalize')
        backend.build()
    
    def download(self, projectDir):
        tool='quartus'
        top_name = self.name
        
        files=[]
        parameters={
              'cable': {'datatype': 'string', 'paramtype': 'generic', 'default':'DE-SoC'},
              'board_device_index': {'datatype': 'string', 'paramtype': 'generic', 'default':'2'},
              }

        edam = {
          'files'        : files,
          'name'         : top_name,
          'parameters'   : parameters,
          'toplevel'     : top_name,
        }        

        backend = edatool.get_edatool(tool)(edam=edam, work_root=projectDir, verbose=True)

        print('Download bitstream')
        backend.run()

    def getInputKey(self):
        key = self.wire('KEY', 4)
        self.addIn('KEY', key)
        keyn = self.wire('KEYn', 4)
        py4hw.Not(self, 'KEYn', key, keyn)
        return keyn
        
    def getOutputHex(self, i):
        assert(i >= 0 and i <= 5)
        name = 'HEX{}'.format(i)
        namen = 'HEXn{}'.format(i)
        p = self.wire(name, 7)
        pn = self.wire(namen, 7)
        
        py4hw.Not(self, namen, pn, p)
        self.addOut(name, p)
        return pn
    
    def getI2C(self):
        #set_location_assignment PIN_J12 -to FPGA_I2C_SCLK
        #set_location_assignment PIN_K12 -to FPGA_I2C_SDAT
        fpga_i2c_sclk = self.addOut('FPGA_I2C_SCLK', self.wire('FPGA_I2C_SCLK'))
        fpga_i2c_sdat = self.addInOut('FPGA_I2C_SDAT', self.wire('FPGA_I2C_SDAT'))
        
        i2c = I2CInterface(self, 'fpga')
        
        py4hw.Buf(self, 'SCL', i2c.SCL, fpga_i2c_sclk)
        py4hw.BidirBuf(self, 'SDA', i2c.SDA_IN, i2c.SDA_OUT, i2c.SDA_OE, fpga_i2c_sdat)
        
        return i2c
    
    def getAudio(self):
        
        audio = AudioInterface(self, 'audio')
        
        self.addInterfaceSource('', audio)
        
        return audio
    
    def getVGAController(self, vga_clk):
        '''
        Connections to the ADV7123
        

        Returns
        -------
        None.

        '''
        vga_ext = VGAExternalControllerInterface(self, 'DE1SoC_EXT_VGA_CTRL')
        vga_int = VGAInternalControllerInterface(self, 'DE1SoC_INT_VGA_CTRL')
        
        self.addInterfaceSource('VGA', vga_ext)
        
        py4hw.Buf(self, 'R', vga_int.R, vga_ext.R)
        py4hw.Buf(self, 'G', vga_int.G, vga_ext.G)
        py4hw.Buf(self, 'B', vga_int.B, vga_ext.B)
        
        py4hw.Buf(self, 'HS', vga_int.HS, vga_ext.HS)
        py4hw.Buf(self, 'VS', vga_int.VS, vga_ext.VS)
        
        py4hw.Buf(self, 'SYNC_N', vga_int.HS, vga_ext.SYNC_N)
        py4hw.Buf(self, 'BLANK_N', vga_int.VS, vga_ext.BLANK_N)
        
        py4hw.Buf(self, 'CLK', vga_clk, vga_ext.CLK)
        
        return vga_int
        
    def getFPGADeviceInfo(self):
        info = {'family':'"Cyclone V"', 'device':'5CSEMA5F31C6', 'package':'FBGA', 'pin_count':'896', 'speed_grade':'6'}
        return info
        
    def getPinAssignments(self):
        pins = {'ADC_CONVST':'PIN_AJ4', 'ADC_DIN':'PIN_AK4', 'ADC_DOUT':'PIN_AK3', 'ADC_SCLK':'PIN_AK2',
                'AUD_ADCDAT':'PIN_K7', 'AUD_ADCLRCK':'PIN_K8', 'AUD_BCLK':'PIN_H7', 'AUD_DACDAT':'PIN_J7', 'AUD_DACLRCK':'PIN_H8', 'AUD_XCK':'PIN_G7',
                'CLOCK2_50':'PIN_AA16', 'CLOCK3_50':'PIN_Y26', 'CLOCK4_50':'PIN_K14', 'CLOCK_50':'PIN_AF14',
                'DRAM_ADDR[0]':'PIN_AK14', 'DRAM_ADDR[1]':'PIN_AH14', 'DRAM_ADDR[2]':'PIN_AG15', 'DRAM_ADDR[3]':'PIN_AE14',
                'DRAM_ADDR[4]':'PIN_AB15', 'DRAM_ADDR[5]':'PIN_AC14', 'DRAM_ADDR[6]':'PIN_AD14', 'DRAM_ADDR[7]':'PIN_AF15',
                'DRAM_ADDR[8]':'PIN_AH15', 'DRAM_ADDR[9]':'PIN_AG13', 'DRAM_ADDR[10]':'PIN_AG12', 'DRAM_ADDR[11]':'PIN_AH13',
                'DRAM_ADDR[12]':'PIN_AJ14', 'DRAM_BA[0]':'PIN_AF13', 'DRAM_BA[1]':'PIN_AJ12', 'DRAM_CAS_N':'PIN_AF11',
                'DRAM_CKE':'PIN_AK13', 'DRAM_CLK':'PIN_AH12', 'DRAM_CS_N':'PIN_AG11',
                'DRAM_DQ[0]':'PIN_AK6', 'DRAM_DQ[1]':'PIN_AJ7', 'DRAM_DQ[2]':'PIN_AK7', 'DRAM_DQ[3]':'PIN_AK8', 
                'DRAM_DQ[4]':'PIN_AK9', 'DRAM_DQ[5]':'PIN_AG10', 'DRAM_DQ[6]':'PIN_AK11', 'DRAM_DQ[7]':'PIN_AJ11',
                'DRAM_DQ[8]':'PIN_AH10', 'DRAM_DQ[9]':'PIN_AJ10', 'DRAM_DQ[10]':'PIN_AJ9', 'DRAM_DQ[11]':'PIN_AH9',
                'DRAM_DQ[12]':'PIN_AH8', 'DRAM_DQ[13]':'PIN_AH7', 'DRAM_DQ[14]':'PIN_AJ6', 'DRAM_DQ[15]':'PIN_AJ5',
                'DRAM_LDQM':'PIN_AB13', 'DRAM_RAS_N':'PIN_AE13', 'DRAM_UDQM':'PIN_AK12', 'DRAM_WE_N':'PIN_AA13',
                'FPGA_I2C_SCLK':'PIN_J12', 'FPGA_I2C_SDAT':'PIN_K12',
                'HEX0[0]':'PIN_AE26', 'HEX0[1]':'PIN_AE27', 'HEX0[2]':'PIN_AE28', 'HEX0[3]':'PIN_AG27', 'HEX0[4]':'PIN_AF28', 
                'HEX0[5]':'PIN_AG28', 'HEX0[6]':'PIN_AH28',
                'HEX1[0]':'PIN_AJ29', 'HEX1[1]':'PIN_AH29', 'HEX1[2]':'PIN_AH30', 'HEX1[3]':'PIN_AG30', 'HEX1[4]':'PIN_AF29', 
                'HEX1[5]':'PIN_AF30', 'HEX1[6]':'PIN_AD27', 
                'HEX2[0]':'PIN_AB23', 'HEX2[1]':'PIN_AE29', 'HEX2[2]':'PIN_AD29', 'HEX2[3]':'PIN_AC28', 'HEX2[4]':'PIN_AD30', 
                'HEX2[5]':'PIN_AC29', 'HEX2[6]':'PIN_AC30', 
                'HEX3[0]':'PIN_AD26', 'HEX3[1]':'PIN_AC27', 'HEX3[2]':'PIN_AD25', 'HEX3[3]':'PIN_AC25', 'HEX3[4]':'PIN_AB28', 
                'HEX3[5]':'PIN_AB25', 'HEX3[6]':'PIN_AB22', 
                'HEX4[0]':'PIN_AA24', 'HEX4[1]':'PIN_Y23', 'HEX4[2]':'PIN_Y24', 'HEX4[3]':'PIN_W22', 'HEX4[4]':'PIN_W24', 
                'HEX4[5]':'PIN_V23', 'HEX4[6]':'PIN_W25', 
                'HEX5[0]':'PIN_V25', 'HEX5[1]':'PIN_AA28', 'HEX5[2]':'PIN_Y27', 'HEX5[3]':'PIN_AB27', 'HEX5[4]':'PIN_AB26',
                'HEX5[5]':'PIN_AA26', 'HEX5[6]':'PIN_AA25',
                'IRDA_RXD':'PIN_AA30', 'IRDA_TXD':'PIN_AB30',
                'KEY[0]':'PIN_AA14', 'KEY[1]':'PIN_AA15', 'KEY[2]':'PIN_W15', 'KEY[3]':'PIN_Y16',
                'LEDR[0]':'PIN_V16', 'LEDR[1]':'PIN_W16', 'LEDR[2]':'PIN_V17', 'LEDR[3]':'PIN_V18', 'LEDR[4]':'PIN_W17', 
                'LEDR[5]':'PIN_W19', 'LEDR[6]':'PIN_Y19', 'LEDR[7]':'PIN_W20', 'LEDR[8]':'PIN_W21', 'LEDR[9]':'PIN_Y21',
                'PS2_CLK':'PIN_AD7', 'PS2_CLK2':'PIN_AD9', 'PS2_DAT':'PIN_AE7', 'PS2_DAT2':'PIN_AE9',
                'SW[0]':'PIN_AB12', 'SW[1]':'PIN_AC12', 'SW[2]':'PIN_AF9', 'SW[3]':'PIN_AF10', 'SW[4]':'PIN_AD11', 
                'SW[5]':'PIN_AD12', 'SW[6]':'PIN_AE11', 'SW[7]':'PIN_AC9', 'SW[8]':'PIN_AD10', 'SW[9]':'PIN_AE12', 
                'TD_CLK27':'PIN_H15', 'TD_DATA[0]':'PIN_D2', 'TD_DATA[1]':'PIN_B1', 'TD_DATA[2]':'PIN_E2', 'TD_DATA[3]':'PIN_B2', 
                'TD_DATA[4]':'PIN_D1', 'TD_DATA[5]':'PIN_E1', 'TD_DATA[6]':'PIN_C2', 'TD_DATA[7]':'PIN_B3', 'TD_HS':'PIN_A5', 
                'TD_RESET_N':'PIN_F6', 'TD_VS':'PIN_A3',
                'VGA_BLANK_N':'PIN_F10', 'VGA_B[0]':'PIN_B13', 'VGA_B[1]':'PIN_G13', 'VGA_B[2]':'PIN_H13', 
                'VGA_B[3]':'PIN_F14', 'VGA_B[4]':'PIN_H14', 'VGA_B[5]':'PIN_F15', 'VGA_B[6]':'PIN_G15', 'VGA_B[7]':'PIN_J14', 
                'VGA_CLK':'PIN_A11', 'VGA_G[0]':'PIN_J9', 'VGA_G[1]':'PIN_J10', 'VGA_G[2]':'PIN_H12', 'VGA_G[3]':'PIN_G10', 
                'VGA_G[4]':'PIN_G11', 'VGA_G[5]':'PIN_G12', 'VGA_G[6]':'PIN_F11', 'VGA_G[7]':'PIN_E11', 'VGA_HS':'PIN_B11', 
                'VGA_R[0]':'PIN_A13', 'VGA_R[1]':'PIN_C13', 'VGA_R[2]':'PIN_E13', 'VGA_R[3]':'PIN_B12', 'VGA_R[4]':'PIN_C12', 
                'VGA_R[5]':'PIN_D12', 'VGA_R[6]':'PIN_E12', 'VGA_R[7]':'PIN_F13', 'VGA_SYNC_N':'PIN_C10', 'VGA_VS':'PIN_D11',
                
                }
        return pins
        
    def getQsf(self):
        info = self.getFPGADeviceInfo()
        qsf = '# build by py4hw\n'
        qsf += 'set_global_assignment -name FAMILY {}\n'.format(info['family'])
        qsf += 'set_global_assignment -name DEVICE {}\n'.format(info['device'])
        qsf += 'set_global_assignment -name DEVICE_FILTER_PACKAGE {}\n'.format(info['package'])
        qsf += 'set_global_assignment -name DEVICE_FILTER_PIN_COUNT {}\n'.format(info['pin_count'])
        qsf += 'set_global_assignment -name DEVICE_FILTER_SPEED_GRADE {}\n'.format(info['speed_grade'])

        pins = self.getPinAssignments()
        
        for key in pins.keys():
            qsf += 'set_location_assignment {} -to {}\n'.format(pins[key], key)

        # by default all are 3.3 LVTTL
        for key in pins.keys():
            qsf += 'set_instance_assignment -name IO_STANDARD "3.3-V LVTTL" -to {}\n'.format(key)
        
        return qsf



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 21 17:40:06 2023

@author: dcr
"""

import py4hw.external.intel.BlockDesignFile
import py4hw.external.intel.VectorWaveformFile
import py4hw.external.intel.ip
from py4hw.external.intel.ip.alt import *
from py4hw.external.intel.ip.lpm import *